% LogicE.lhs
%
% This file is part of CalcCheck
%
% Copyright (c) 2011  Wolfram Kahl
% All rights reserved
%
\section{|LogicE|: The Equational Logic E}

%{{{ module LogicE
\begin{code}
{-# LANGUAGE FlexibleInstances, PatternGuards #-}
module LogicE where

import qualified AC
import TeX
import Utils.List (groupFst, uninsert)
import Options (CalcCheckOpts(), programId, programIdPrefix, programIdPrefixHTML)

import Data.Map (Map)
import qualified Data.Map as Map

import Data.Set (Set)
import qualified Data.Set as Set

import Data.Function (on)
import Data.Maybe (mapMaybe)
import Data.List (nub, foldl', groupBy, partition)
import qualified Data.Char as Char

import Control.Arrow (second)
import Control.Monad (liftM, liftM2)
import Control.Monad.Error ()
\end{code}
%}}}

%{{{ \subsection{Expressions}
\subsection{Expressions}

\begin{code}
type Var = String
\end{code}

\begin{code}
freshVars :: [Var]
freshVars = map (('v' :) . show) [1 :: Integer ..]
\end{code}

%{{{ type Op = Char
\begin{code}
type Op = Char
\end{code}
%}}}

\begin{code}
data Expr
  = Var Var
  | IntLit Integer
  | Const Op
  | Un Op Expr
  | Bin Op Expr Expr
  | VarArgs Op [Expr]
  | TypeAnn Expr Expr
  | ConjChain Expr [(Op, Expr)]
  | Subst Expr [Var] [Expr]
  | Quant Op [(Var, Maybe Expr)] (Maybe Expr) Expr
  | SetEnum [Expr]
  | SetCompr [(Var, Maybe Expr)] (Maybe Expr) (Maybe Expr)
\end{code}

\begin{code}
-- |instance Show Expr where showsPrec p = showsPrec p . toTerm|
\end{code}

%{{{ unBin, unUn, unConst
|unBin| serves to extract the operator character
from operator parsers, and is used only by parsers
that expects isolated operators
(i.e., operators not occurring as expression constructors),
in particular in the first argument of \verb|\CalcStep|.
Formally, |unBin| is the least function satisfying
|unBin (Bin op e1 e2) = op|:
\begin{code}
unBin :: (Expr -> Expr -> Expr) -> Op
unBin f = case f undefined undefined of
  Bin op _ _ -> op
  _ -> error "unBin"
\end{code}
Similarly, |unUn| is specified by |unUn (Un op e) = op|:
\begin{code}
unUn :: (Expr -> Expr) -> Op
unUn f = case f undefined of
  Un op _ -> op
  _ -> error "unUn"
\end{code}
And |unConst| is the left-inverse of |Const|,
that is, |unConst . Const = id|,
or |unConst (Const op) = op|:
\begin{code}
unConst :: Expr -> Op
unConst f = case f of
  Const op -> op
  _ -> error "unConst"
\end{code}
%}}}

%{{{ unchain
For unfolding a |ConjChain| into its constituent conjunctions:
\begin{code}
unChain :: e -> [(op,e)] -> [(op,(e, e))]
unChain _ [] = error "unChain: empty ConjChain"
unChain e ((op, e') : ps) = (op, (e, e')) : h e' ps
  where
    h _ [] = []
    h t ((o,t') : qs) = (o, (t, t')) : h t' qs
\end{code}
%}}}

\begin{code}
data Const = Int Integer
  deriving (Eq, Ord)

showsConst :: Const -> ShowS
showsConst (Int i) = shows i

instance Show Const where showsPrec _ = showsConst
\end{code}

%{{{ toTerm :: Expr -> AC.Term Var Op
\edcomm{WK}{|toTerm| needs to use fresh variables for bound variables!}

\begin{code}
binTerm :: (Ord var, Ord cnst) => Op -> AC.Term var Op cnst -> AC.Term var Op cnst -> AC.Term var Op cnst
binTerm op t u  | isAC op    =  AC.applyAC op t u 
                | isC op     =  AC.applyC op t u 
                | otherwise  =  AC.apply op [t,u]

toTerm :: Expr -> AC.Term Var Op Const
toTerm (Var v) = AC.var v
toTerm (IntLit i) = AC.Const (Int i)
toTerm (Const op) = AC.apply op []
toTerm (Un op e) = AC.apply op [toTerm e]
toTerm (Bin op e1 e2) = binTerm op (toTerm e1) (toTerm e2)
toTerm (VarArgs op es) = AC.apply op $ map toTerm es
toTerm (TypeAnn e _t) = toTerm e -- ignoring type annotations in matching and rewriting
toTerm (ConjChain _ []) = error "toTerm: internal error: empty ConjChain!"
toTerm (ConjChain e ps) = foldr1 (binTerm '∧') .
   map (\ (o,(t, t')) -> binTerm o t t') $ unChain (toTerm e) (map (second toTerm) ps)
toTerm (Subst e vs es) = AC.apply '⇇' (map AC.var vs ++ map toTerm (es ++ [e]))
toTerm (Quant op vs mr e) = AC.apply (maybe op id $ binToQuant op)
  [AC.applyACs '∾' (map (AC.var . fst) vs), toTerm (maybe true id mr), toTerm e]
toTerm (SetEnum []) = AC.apply '∅' []
toTerm (SetEnum es) = foldr1 (binTerm '⊍') (map (\ e -> AC.apply '︷' [toTerm e]) es)
toTerm (SetCompr vs mr Nothing) = toTerm . SetCompr vs mr . Just . tuple $ map (Var . fst) vs
toTerm (SetCompr vs mr (Just e)) = toTerm $ Quant '⨃' vs mr $ Un '︷' e
\end{code}

\begin{code}
unVarList :: (Show var, Show cnst) => AC.Term var Op cnst -> [var]
unVarList (AC.Var v) = [v]
unVarList (AC.AC '∾' ts) = ts >>= unVarList
unVarList e = error $ "unVarList: non-variable encountered: " ++ AC.showsTerm e "!"
\end{code}
%}}}
%}}}

%{{{ \subsection{Free Variables and $\occurs{}{}$}
\subsection{Free Variables and $\occurs{}{}$}

%{{{ freeVars :: Expr -> Set Var
\begin{code}
freeVarss :: [Expr] -> Set Var
freeVarss = Set.unions . map freeVars

freeVarsm :: Maybe Expr -> Set Var
freeVarsm = maybe Set.empty freeVars

freeVars :: Expr -> Set Var
freeVars (Var v) = Set.singleton v
freeVars (IntLit _) = Set.empty
freeVars (Const _) = Set.empty
freeVars (Un _ e) = freeVars e
freeVars (Bin _ e1 e2) = freeVars e1 `Set.union` freeVars e2
freeVars (VarArgs _ es) = freeVarss es
freeVars (TypeAnn e t) = freeVars e `Set.union` freeVars t
freeVars (ConjChain e ps) = freeVarss $ e : map snd ps
freeVars (Subst e vs es) =  freeVarss es `Set.union`
                            (freeVars e `Set.difference` Set.fromList vs)
freeVars (Quant _op vs mr e) = freeVarss (mapMaybe snd vs)  `Set.union`
  ((maybe id (Set.union . freeVars) mr $ freeVars e) `Set.difference` Set.fromList (map fst vs))
freeVars (SetEnum es) = freeVarss es
freeVars (SetCompr vs mr me) = freeVarss (mapMaybe snd vs)  `Set.union`
  ((maybe id (Set.union . freeVars) mr $ freeVarsm me) `Set.difference` Set.fromList (map fst vs))
\end{code}
%}}}

\begin{code}
occurs :: [Var] -> [Expr] -> Bool
occurs vs es = any (`Set.member` f) vs
  where f = freeVarss es


cleanOccsNOccs  :: [([( Var, AC.Term Var Op Const)], [(Var, AC.Term Var Op Const)])]
                -> ([( Var, AC.Term Var Op Const)], [(Var, AC.Term Var Op Const)])
cleanOccsNOccs ps = let
  (occs1, oks1) = unzip ps
  reallyOccurs (v, t) = occursInTerm [v] [t]
  (occs2, oks2) = partition reallyOccurs (concat occs1)
  in (occs2 , nub (concat (oks2 : oks1)))
\end{code}

%{{{ freeVarsT :: AC.Term Var Op Const -> Set Var
\begin{code}
freeVarssT :: [AC.Term Var Op Const] -> Set Var
freeVarssT = Set.unions . map freeVarsT

freeVarsmT :: Maybe (AC.Term Var Op Const) -> Set Var
freeVarsmT = maybe Set.empty freeVarsT

freeVarsT :: AC.Term Var Op Const -> Set Var
freeVarsT (AC.Var v) = Set.singleton v
freeVarsT (AC.Const _) = Set.empty
freeVarsT t@(AC.Appl '⇇' ts) = case splitSubstArgs ts of
  Left e -> error $ "freeVarsT: " ++ shows t e
  Right (e, (es, vs)) -> freeVarssT es `Set.union`
                            (freeVarsT e `Set.difference` Set.fromList (mapMaybe AC.unVar vs))
freeVarsT (AC.Appl _op (vs@(AC.AC '∾' _) : ts)) = -- quantification bound variables
  freeVarssT ts `Set.difference` Set.fromList (unVarList vs)
freeVarsT (AC.Appl op (vs : ts)) | Just _bin <- quantToBin op = -- quantification op
  freeVarssT ts `Set.difference` Set.fromList (unVarList vs)
freeVarsT (AC.Appl _ ts) = freeVarssT ts
freeVarsT (AC.C _ t1 t2) = freeVarssT [t1, t2]
freeVarsT (AC.AC _ ts) = freeVarssT ts
\end{code}
%}}}

%{{{ occursInTerm
\begin{code}
occursInTerm :: [Var] -> [AC.Term Var Op Const] -> Bool
occursInTerm vs ts = any (`Set.member` f) vs
  where f = freeVarssT ts
\end{code}
%}}}

%{{{ freeVarsWithScope :: Expr -> Set (Var, Set Var)
For variable capture considerations, it is important to know
for each free variable,
which binders are in scope at its occurrences.
\begin{code}
freeVarssWithScope :: [Expr] -> Set (Var, Set Var)
freeVarssWithScope = Set.unions . map freeVarsWithScope

freeVarsmWithScope :: Maybe Expr -> Set (Var, Set Var)
freeVarsmWithScope = maybe Set.empty freeVarsWithScope

freeVarsWithScope :: Expr -> Set (Var, Set Var)
freeVarsWithScope (Var v) = Set.singleton (v, Set.empty)
freeVarsWithScope (IntLit _) = Set.empty
freeVarsWithScope (Const _) = Set.empty
freeVarsWithScope (Un _ e) = freeVarsWithScope e
freeVarsWithScope (Bin _ e1 e2) = freeVarsWithScope e1 `Set.union` freeVarsWithScope e2
freeVarsWithScope (VarArgs _ es) = freeVarssWithScope es
freeVarsWithScope (TypeAnn e t) = freeVarsWithScope e `Set.union` freeVarsWithScope t
freeVarsWithScope (ConjChain e ps) = freeVarssWithScope $ e : map snd ps
freeVarsWithScope (Subst e vs es) =  freeVarssWithScope es `Set.union`
   (Set.map  (\ (v,bs) -> (v, Set.union bs vS))
             (Set.filter ((`Set.notMember` vS) . fst) $ freeVarsWithScope e))
  where vS = Set.fromList vs
freeVarsWithScope (Quant _op vs mr e) = freeVarssWithScope (mapMaybe snd vs)  `Set.union`
   (Set.map  (\ (v,bs) -> (v, Set.union bs vS))
             (Set.filter ((`Set.notMember` vS) . fst) 
                         (maybe id (Set.union . freeVarsWithScope) mr $ freeVarsWithScope e)))
  where vS = Set.fromList (map fst vs)
freeVarsWithScope (SetEnum es) = freeVarssWithScope es
freeVarsWithScope (SetCompr vs mr me) = freeVarssWithScope (mapMaybe snd vs)  `Set.union`
  (Set.map  (\ (v,bs) -> (v, Set.union bs vS))
            (Set.filter ((`Set.notMember` vS) . fst) 
                        (maybe id (Set.union . freeVarsWithScope) mr $ freeVarsmWithScope me)))
  where vS = Set.fromList (map fst vs)
\end{code}
%}}}

%{{{ scopeFromExpr
A different view collects all un-dominated free variables
of every scope together, and inner scopes with their binders.
\begin{code}
data Scope = Scope
  {scopeFree :: Set Var
  ,scopeInner :: Set (Set Var, Scope)  -- bvars and inner scope
  }
  deriving (Eq, Ord)

emptyScope :: Scope
emptyScope = Scope Set.empty Set.empty

innerScope :: Set Var -> Scope -> Scope
innerScope vs s = Scope Set.empty (Set.singleton (vs, s))

scopeJoin :: Scope -> Scope -> Scope
scopeJoin s1 s2 = Scope
  {scopeFree = Set.union (scopeFree s1) (scopeFree s2)
  ,scopeInner = Set.union (scopeInner s1) (scopeInner s2)
  }

scopeJoins :: [Scope] -> Scope
scopeJoins = foldr scopeJoin emptyScope

scopeFromExprs :: [Expr] -> Scope
scopeFromExprs = scopeJoins . map scopeFromExpr

scopeFromExprM :: Maybe Expr -> Scope
scopeFromExprM = maybe emptyScope scopeFromExpr

scopeFromExpr :: Expr -> Scope
scopeFromExpr (Var v) = Scope (Set.singleton v) Set.empty
scopeFromExpr (IntLit _) = emptyScope
scopeFromExpr (Const _) = emptyScope
scopeFromExpr (Un _ e) = scopeFromExpr e
scopeFromExpr (Bin _ e1 e2) = scopeFromExpr e1 `scopeJoin` scopeFromExpr e2
scopeFromExpr (VarArgs _ es) = scopeFromExprs es
scopeFromExpr (TypeAnn e t) = scopeFromExpr e `scopeJoin` scopeFromExpr t
scopeFromExpr (ConjChain e ps) = scopeFromExprs $ e : map snd ps
scopeFromExpr (Subst e vs es) =  scopeFromExprs es `scopeJoin`
   innerScope vS (scopeFromExpr e)
  where vS = Set.fromList vs
scopeFromExpr (Quant _op vs mr e) = scopeFromExprs (mapMaybe snd vs)  `scopeJoin`
   innerScope vS (scopeFromExpr e `scopeJoin` scopeFromExprM mr)
  where vS = Set.fromList (map fst vs)
scopeFromExpr (SetEnum es) = scopeFromExprs es
scopeFromExpr (SetCompr vs mr me) = scopeFromExprs (mapMaybe snd vs)  `scopeJoin`
   innerScope vS (scopeFromExprM me `scopeJoin` scopeFromExprM mr)
  where vS = Set.fromList (map fst vs)
\end{code}
%}}}

\begin{code}
data NotOccurs = NotOccurs
  { notOccVars :: [Var]
  , notOccExprs :: [Expr]
  }
\end{code}

\begin{code}
showsOccurs, showsNotOccurs :: ShowS -> ShowS -> ShowS
showsOccurs v e = ("occurs(‘" ++)  . v . ("’, ‘" ++) . e . ("’)" ++)
showsNotOccurs v e = ('¬' :) .  showsOccurs v e

instance Show NotOccurs where
  showsPrec _ no  =  showsNotOccurs
                       (showsCommaSep1 (map showsVar $ notOccVars no))
                       (showsCommaSep1 (map shows $ notOccExprs no))
\end{code}

\begin{code}
notOccursToAC :: [NotOccurs] -> [AC.NotOccurs Var Op Const]
notOccursToAC nos = let
  doNO (NotOccurs vs es) = do
    v <- vs
    e <- es
    [(toTerm e, v)]
  in map (\ (e, vs) -> (nub vs, e)) $ groupFst (nos >>= doNO)
\end{code}
%}}}

%{{{ \subsection{Variable Capture Provisos}
\subsection{Variable Capture Provisos}

If a free variable $R$ occurs below a binder for $x$ in one place,
but outside any binder for $x$ in another place,
we infer the proviso $\lnot\occurs{x}{R}$.

\begin{code}
proviso :: Expr -> [NotOccurs]
proviso e = let
  vss :: [(Var, [Set Var])]
  vss = groupFst $ Set.toAscList (freeVarsWithScope e)
  vs = Set.fromList $ map fst vss
  f :: (Var, [Set Var]) -> [NotOccurs]
  f (_fv, []) = []
  f (_fv, [_]) = []
  f (fv, bss) = case Set.toAscList . Set.unions $ do
      (bs1, bss') <- uninsert bss
      bs2 <- bss'
      [Set.difference bs1 bs2 `Set.union` Set.difference bs2 bs1]
    of
      [] -> []
      bs -> if all (`Set.member` vs) bs
            then [] -- global variables don't count, see (11.7)
            else [NotOccurs (filter (`Set.notMember` vs) bs) [Var fv]]
  in vss >>= f
\end{code}

For checking whether we infer the right provisos:
\begin{code}
listProvisos :: IO ()
listProvisos = let
    thmsP = quantLawsP ++ predLawsP
    thms = substLaws ++ quantLaws ++ predLaws ++ setLaws ++ relLaws
    rP [] = "-- no proviso?\n"
    rP nos = showsCommaSep1 (map shows nos) "\n"
    r [] = ""
    r nos = "Proviso!!!   " ++ showsCommaSep1 (map shows nos) "\n"
    f g thm = do
      putStrLn $ showTheorem thm
      putStrLn $ replicate 10 ' ' ++ g (proviso (snd thm))
  in mapM_ (f r) thms >> mapM_ (f rP) thmsP
\end{code}
%}}}

%{{{ \subsection{Laws Management}
\subsection{Laws Management}

\begin{code}
type Theorem = (([String], [String]), Expr)
\end{code}

\begin{code}
showsTheorem :: Theorem -> ShowS
showsTheorem ((ns, ds),e)  = (case ns of [] -> id; n : _ -> (n ++) . (' ' :))
                           . (foldr (++) `flip` take 1 ds)
                           .  (' ' :) . shows e

showTheorem :: Theorem -> String
showTheorem = flip showsTheorem ""
\end{code}

\begin{code}
mkRules :: [(t, Expr)] -> [((AC.TermPair Var Op Const, [AC.NotOccurs Var Op Const]), (t, Expr))]
mkRules = foldr h []
  where
    h r@(_descr, e) rs | Just _op <- isAssocLaw e = let v = AC.var "p" in (((v,v),[]),r) : rs
    h r@(_descr, e) rs = foldr (\ eq -> (((eq, nos), r) :)) rs (equations' =<< equEquiv (toTerm e))
      where nos = notOccursToAC $ proviso e
\end{code}

\begin{code}
allRules :: [((AC.TermPair Var Op Const, [AC.NotOccurs Var Op Const]), Theorem)]
allRules = mkRules laws
\end{code}

\begin{code}
equations' :: AC.Term Var Op Const -> [AC.TermPair Var Op Const]
equations' t = AC.nubEqus LogicE.freshVars . falseEqu t . andEqu t $ 
  (t, toTerm true) :
  -- |(toTerm true, t) :|
  equations t
\end{code}

A law $\lnot t$ is automatically turned into an equation $t = \false$:
\begin{code}
falseEqu :: AC.Term Var Op Const -> [AC.TermPair Var Op Const] -> [AC.TermPair Var Op Const]
falseEqu (AC.Appl '¬' [t]) = ((t, toTerm false) :)
falseEqu _ = id
\end{code}

A conjunction law (e.g. (15.44) Trichotomy) automatically contributes
its constitutent laws:

\begin{code}
andEqu :: AC.Term Var Op Const -> [AC.TermPair Var Op Const] -> [AC.TermPair Var Op Const]
andEqu (AC.AC '∧' ts) r = foldr (\ t r' -> equations' t ++ r') r ts
andEqu _ r = r
\end{code}

Equalities in laws can be instantiated as equivalences
\edcomm{WK}{only if typing allows}
\edcomm{WK}{inefficient! (gives all possibilities to replace a subset of equalities with equivalences)}:
\begin{code}
equEquiv :: AC.Term Var Op Const -> [AC.Term Var Op Const]
equEquiv t@(AC.Var _) = [t]
equEquiv t@(AC.Const _) = [t]
equEquiv (AC.AC op ts) = fmap (AC.applyACs op) $ mapM equEquiv ts
equEquiv (AC.Appl op ts) = fmap (AC.apply op) $ mapM equEquiv ts
equEquiv (AC.C op t1 t2) = do
  t1' <- equEquiv t1
  t2' <- equEquiv t2
  AC.applyC op t1 t2 : case op of
    '=' -> [AC.applyAC '≡' t1' t2']
    _ -> []
\end{code}

\begin{code}
equations :: AC.Term Var Op Const -> [AC.TermPair Var Op Const]
equations (AC.C '=' t u) = [(t,u),(u,t)]
equations (AC.AC '≡' ts) = map (\ (t,u) -> (AC.mkAC '≡' t, AC.mkAC '≡' u)) $ nub $ splits ts
equations _ = []
\end{code}

\begin{code}
splits :: [t] -> [([t], [t])]
splits [] = []
splits [_] = []
splits (x : xs) = ([x], xs) : (xs, [x]) : do
   (ys, zs) <- splits xs
   [(x : ys, zs),(ys, x : zs)]
\end{code}

\begin{code}
idLaw :: Expr
idLaw = Var "p" ≡ Var "p"
\end{code}

\begin{code}
isSymLaw :: Expr -> Maybe Op
isSymLaw e = case e of
  Bin op (Bin op1 (Var x) (Var y)) (Bin op2 (Var y') (Var x'))
   | op `elem` "≡=" && op1 == op2 && x == x' && y == y' && x /= y
   -> Just op1
  _ -> Nothing
\end{code}

\begin{code}
getSymLaws :: [(t, Expr)] -> [(Op, (t, Expr))]
getSymLaws = mapMaybe (\ law@(_, e) -> fmap (\ p -> (p , law)) $ isSymLaw e)

symLaws :: [(Op, Theorem)]
symLaws = getSymLaws laws
\end{code}

We exclude associativity laws (i.e., where |op1 == op2|)
from the results of |isMutAssocLaw|:
\begin{code}
isMutAssocLaw :: Expr -> Maybe (Op , Op)
isMutAssocLaw e = case e of
  Bin op e1 e2 | op `elem` "≡=" -> case isLeft e1 of
    Just q -> isRight q e2
    Nothing -> case isLeft e2 of
      Just q -> isRight q e1
      Nothing -> Nothing
  _ -> Nothing
  where
    isLeft (Bin op2 (Bin op1 (Var x) (Var y)) (Var z))
      | op1 /= op2 && length (nub [x,y,z]) == 3  = Just (op1,op2,x,y,z)
    isLeft _ = Nothing

    isRight (op1, op2, x,y,z) (Bin op1' (Var x') (Bin op2' (Var y') (Var z')))
      | op1 == op1' && op2 == op2' && x == x' && y == y' && z == z'
      = Just (op1, op2)
    isRight _ _ = Nothing
\end{code}

\begin{code}
getMutAssocLaws0, getMutAssocLaws :: [(t, Expr)] -> [((Op, Op), (t, Expr))]
getMutAssocLaws0 = mapMaybe (\ law@(_, e) -> fmap (\ p -> (p , law)) $ isMutAssocLaw e)
getMutAssocLaws thms = foldr f [] (getMutAssocLaws0 thms)
  where
    syms = map fst $ getSymLaws thms
    f p@((op1, op2), thm) r = p :  if op1 `elem` syms && op2 `elem` syms
                                   then ((op2, op1), thm) : r
                                   else r

mutAssocLaws :: [((Op, Op), Theorem)]
mutAssocLaws = getMutAssocLaws laws
\end{code}

\begin{code}
isAssocLaw :: Expr -> Maybe Op
isAssocLaw e = case e of
  Bin op e1 e2 | op `elem` "≡=" -> case isLeft e1 of
    Just q -> isRight q e2
    Nothing -> case isLeft e2 of
      Just q -> isRight q e1
      Nothing -> Nothing
  _ -> Nothing
  where
    isLeft (Bin op1 (Bin op2 (Var x) (Var y)) (Var z))
      | op1 == op2 && length (nub [x,y,z]) == 3  = Just (op1,x,y,z)
    isLeft _ = Nothing

    isRight (op, x,y,z) (Bin op1 (Var x') (Bin op2 (Var y') (Var z')))
      | op == op1 && op == op2 && x == x' && y == y' && z == z'
      = Just op
    isRight _ _ = Nothing
\end{code}
%}}}

%{{{ \subsection{Calculations}
\subsection{Calculations}

\begin{code}
data Report = Report
  { message :: [String]
  , confirmedOK :: Bool
  }

prependReport :: String -> Report -> Report
prependReport s r = r
  { message = case message r of
      [] -> [s]
      x : xs -> (s ++ x) : xs
  }

concatReport :: Report -> Report -> Report
concatReport r1 r2 = Report
  { message = message r1 ++ message r2
  , confirmedOK = confirmedOK r1 && confirmedOK r2
  }

concatMReport :: Maybe Report -> Maybe Report -> Maybe Report
concatMReport Nothing m2 = m2
concatMReport m1@(Just _) Nothing = m1
concatMReport (Just r1) (Just r2) = Just (concatReport r1 r2)
\end{code}

\begin{code}
data CalcStep = CalcStep
  { calcOp :: Op
  , calcHint :: [TeX]
  , calcCheck :: Report
  , calcResult :: Expr
  }
\end{code}

\begin{code}
breakAnnot :: Bool
breakAnnot = True

breakA, breakNL :: ShowS
breakA = if breakAnnot then ("\n      " ++) else id
breakNL = if breakAnnot then ('\n' :) else id

fontColor :: String -> ShowS -> ShowS
fontColor rgb s = ("<font color=\"#" ++) . (rgb ++) . ("\">" ++) . s . ("</font>" ++)

ccDarkRed, ccDarkGreen  :: ShowS -> ShowS
ccDarkRed = fontColor "AA0000"
ccDarkGreen = fontColor "00AA00"
\end{code}

\begin{code}
unlines' :: [String] -> String
unlines' [] = ""
unlines' xs = foldr1 (\ ys zs -> ys ++ '\n' : zs) xs
\end{code}

\begin{code}
showsCalcStepReport :: CalcCheckOpts -> Report -> ShowS
showsCalcStepReport cco r = breakA . programIdPrefix cco . flip (foldr opShows) (unlines' $ message r)
\end{code}

\begin{code}
showsCalcStep :: CalcCheckOpts -> CalcStep -> ShowS
showsCalcStep cco (CalcStep op hint check e)
   = ("\n  " ++) . (op :) . ("   〈 " ++) . shows hint . (" 〉" ++) .
     (showsCalcStepReport cco check) .
     ("\n    " ++) . shows e
\end{code}

\begin{code}
showsCalcStepReportHTML :: CalcCheckOpts -> Report -> ShowS
showsCalcStepReportHTML cco r = let
  ss =  breakA . programIdPrefixHTML cco . flip (foldr opHTML) (unlines' $ message r)
  in (if confirmedOK r then ccDarkGreen else ccDarkRed) ss
\end{code}

\begin{code}
showsCalcStepHTML :: CalcCheckOpts -> CalcStep -> ShowS
showsCalcStepHTML cco (CalcStep op hint check e)
   =  ("\n  " ++) . opHTML op . ("   &lang; " ++) . flip (foldr opHTML) (show hint) . (" &rang;" ++) .
      showsCalcStepReportHTML cco check .
      ("\n    " ++) . showsExprHTML e
\end{code}

\begin{code}
data CalcGoal = CalcGoal
  { goalNr    :: Maybe String
  , goalName  :: Maybe String
  , goalExpr  :: Maybe Expr
  }
\end{code}

\begin{code}
noGoal :: CalcGoal
noGoal = CalcGoal Nothing Nothing Nothing
\end{code}

\begin{code}
isNoGoal :: CalcGoal -> Bool
isNoGoal (CalcGoal Nothing Nothing Nothing) = True
isNoGoal _ = False
\end{code}

\begin{code}
showsGoal :: CalcGoal -> ShowS
showsGoal g =
  (case goalNr g of Nothing -> id; Just nr -> ("  (" ++) . (nr ++) . (')' :)) .
  (case goalName g of
     Nothing  ->  id
     Just n   ->  ("  " ++) . flip (foldr opShows) n
              .   case goalExpr g of Nothing -> id; Just _ -> (':' :)
  ) .
  (case goalExpr g of Nothing -> id; Just e -> ("  " ++) . shows e)
\end{code}

\begin{code}
goalAddNr :: String -> CalcGoal -> CalcGoal
goalAddNr s g = g { goalNr = Just s }

goalAddName :: String -> CalcGoal -> CalcGoal
goalAddName s g = g { goalName = Just s }

goalAddExpr :: Expr -> CalcGoal -> CalcGoal
goalAddExpr e g = g { goalExpr = Just e }
\end{code}

\begin{code}
data Calc = Calc
  { calcGoal  :: CalcGoal
  , calcGoalCheck  :: [Report]
  , calcStartDecl :: CalcGoal
  , calcEndDecl :: CalcGoal
  , calcStart :: Expr
  , calcSteps :: [CalcStep]
  }
\end{code}

\begin{code}
showsThisIs :: CalcGoal -> ShowS
showsThisIs g = if isNoGoal g then id else ("   ─ This is: " ++) . showsGoal g
\end{code}

\begin{code}
showsCalcGoalCheckReport :: CalcCheckOpts -> Report -> String
showsCalcGoalCheckReport cco r = programId cco ++ ": " ++ (unlines $ message r)

showsCalcGoalCheck :: CalcCheckOpts -> [Report] -> ShowS
showsCalcGoalCheck _cco [] = id
showsCalcGoalCheck cco rs = ("\n\n" ++) . programIdPrefix cco . ((unlines' $ message =<< rs) ++) . ("\n\n" ++)
\end{code}

\begin{code}
showsCalc :: CalcCheckOpts -> Calc -> ShowS
showsCalc cco c =
   (if isNoGoal $ calcGoal c then id else ("Proving" ++) . showsGoal (calcGoal c) . ("\n\n" ++)) .
   ("    " ++) . shows (calcStart c) . showsThisIs (calcStartDecl c) .
   flip (foldr (showsCalcStep cco)) (calcSteps c) . showsThisIs (calcEndDecl c) .
   showsCalcGoalCheck cco (calcGoalCheck c)

showsCalcs :: CalcCheckOpts -> [Calc] -> ShowS
showsCalcs cco = showsSep  ("\n\n" ++) . map (showsCalc cco)
\end{code}

\begin{code}
showsCalcGoalReportHTML :: CalcCheckOpts -> Report -> ShowS
showsCalcGoalReportHTML cco r = (if confirmedOK r then ccDarkGreen else ccDarkRed)
   (programIdPrefixHTML cco . flip (foldr opHTML) (unlines $ message r))

showsCalcGoalCheckHTML :: CalcCheckOpts -> [Report] -> ShowS
showsCalcGoalCheckHTML _cco [] = id
showsCalcGoalCheckHTML cco rs = ("\n\n" ++) . flip (foldr (showsCalcGoalReportHTML cco)) rs . ("\n\n" ++)
\end{code}

\begin{code}
showsGoalHTML :: CalcGoal -> ShowS
showsGoalHTML g =
  (case goalNr g of Nothing -> id; Just nr -> ("  <b>(" ++) . (nr ++) . (")</b>" ++)) .
  (case goalName g of
     Nothing  ->  id
     Just n   ->  ("  " ++) . flip (foldr opHTML) n
              .   case goalExpr g of Nothing -> id; Just _ -> (':' :)
  ) .
  (case goalExpr g of Nothing -> id; Just e -> ("  " ++) . showsExprHTML e)
\end{code}

\begin{code}
showsThisIsHTML :: CalcGoal -> ShowS
showsThisIsHTML g = if isNoGoal g then id else (" &nbsp; &nbsp;  &mdash; This is: " ++) . showsGoalHTML g
\end{code}

\begin{code}
showsCalcHTML :: CalcCheckOpts -> Calc -> ShowS
showsCalcHTML cco c =
   (if isNoGoal $ calcGoal c then id else ("<strong>Proving</strong>" ++) . showsGoalHTML (calcGoal c) . (":\n\n" ++)) .
   ("    " ++) . showsExprHTML (calcStart c) . showsThisIsHTML (calcStartDecl c) .
   flip (foldr (showsCalcStepHTML cco)) (calcSteps c) . showsThisIsHTML (calcEndDecl c) .
   showsCalcGoalCheckHTML cco (calcGoalCheck c)

showsCalcsHTML :: CalcCheckOpts -> [Calc] -> ShowS
showsCalcsHTML cco = showsSep ("\n\n" ++) . map (showsCalcHTML cco)
\end{code}
%}}}

%{{{ \subsection{Declarations}
\subsection{Declarations}

\begin{code}
data Decl
  = DeclType [Var] Expr (Maybe String)  -- at least one variable
  | DeclEqu Expr Op (Either String Expr) (Maybe String)
\end{code}

%{{{ showsDecl
\begin{code}
showsDecl :: Decl -> ShowS
showsDecl = genShowsDecl
  ('\t' :) -- tab
  showsVar
  opShows
  shows                                     -- showsExpr
  (\ s -> ("⟪ " ++) . (s ++) . (" ⟫" ++))   -- showsInformal
  (\ s -> (" ─ " ++) . (s ++))              -- showsRemark
\end{code}
%}}}

%{{{ showsDeclHTML
\begin{code}
showsDeclHTML :: Decl -> ShowS
showsDeclHTML = genShowsDecl
  ('\t' :) -- tab
  showsVar
  opHTML
  showsExprHTML
  (\ s -> ("<font color=\"#CCCCCC\">" ++) . flip (foldr opHTML) s . ("</font>" ++))   -- showsInformal
  (\ s -> (" &nbsp; &mdash; &nbsp; " ++) . flip (foldr opHTML) s)              -- showsRemark
\end{code}
%}}}

%{{{ genShowsDecl
\begin{code}
genShowsDecl  :: ShowS
              -> (Var -> ShowS)
              -> (Op -> ShowS)
              -> (Expr -> ShowS)
              -> (String -> ShowS)
              -> (String -> ShowS)
              -> Decl -> ShowS
genShowsDecl tab shVar shOp shExpr showsInformal showsRemark d = case d of
  DeclType vs e mr -> showsCommaSep (map shVar vs) . tab . spaced (':' :) . shExpr e . maybe id showsRemark mr
  DeclEqu lhs op rhs mr -> shExpr lhs . tab . spaced ((':' :) . shOp op)
    . either showsInformal shExpr rhs . maybe id showsRemark mr
\end{code}
%}}}
%}}}

%{{{ \subsection{Formal Developments}
\subsection{Formal Developments}

\begin{code}
data Item
  = Calcs [Calc]
  | Decls [Decl]

calcItems :: [Item] -> [[Calc]]
calcItems [] = []
calcItems (Calcs x : items) = x : calcItems items
calcItems (Decls _ : items) = calcItems items
\end{code}

\begin{code}
mapCalcs :: ([Calc] -> [Calc]) -> Item -> Item
mapCalcs f (Calcs cs) = Calcs (f cs)
mapCalcs _ d = d
\end{code}

\begin{code}
showsItem, showsItemHTML :: CalcCheckOpts -> Item -> ShowS
showsItem cco (Calcs cs) = showsCalcs cco cs
showsItem _cco (Decls ds) = showsSep ("\n\n" ++) $ map showsDecl ds
showsItemHTML cco (Calcs cs) = showsCalcsHTML cco cs
showsItemHTML _cco (Decls ds) = showsSep ("\n\n" ++) $ map showsDeclHTML ds
\end{code}

\begin{code}
okCalc :: Calc -> Bool
okCalc c = all confirmedOK (calcGoalCheck c)

okCalcs :: [Calc] -> [Report]
okCalcs cs = let
    n = length cs
    outOf k s =  if n == 1 then "1 calculation" ++ s
                 else shows k " out of " ++ shows n " calculations" ++ s
    (csOK , csNOK) = partition okCalc cs
  in (case csOK of
       [] -> []
       _ -> [Report [outOf (length csOK) " fully checked"] True]
     ) ++ (case csNOK of
       [] -> []
       _ -> [Report [outOf (length csNOK) " not fully justified"] False]
     )
\end{code}

\begin{code}
itemsSummary :: [Item] -> [Report]
itemsSummary items = okCalcs . concat $ calcItems items

itemsSummary' :: CalcCheckOpts -> [Item] -> String
itemsSummary' cco items = showsCalcGoalCheckReport cco =<< itemsSummary items
\end{code}

\begin{code}
showsItems, showsItemsHTML :: CalcCheckOpts -> [Item] -> ShowS
showsItems cco = showsSep ("\n\n=======================\n\n" ++) . map (showsItem cco)
showsItemsHTML cco items = showsSep ("\n<hr/>\n" ++) (map (wrapItemHtml . showsItemHTML cco) items ++ [wrapItemHtml $ showsCalcGoalCheckHTML cco (itemsSummary items)])
\end{code}
\begin{verbatim}
{-
showsItemsHTML is = let
  s =  showsSep ("\n</pre>\n<hr/>\n<pre>\n" ++) . map showsItemHTML
-}
\end{verbatim}

\begin{code}
wrapItemHtml :: ShowS -> ShowS
-- |wrapItemHtml s = ("\n<pre>\n" ++) . s . ("\n</pre>\n" ++)|

wrapItemHtml s = f (s "")
  where
    f (' ' : cs) = ("&nbsp;" ++) . f cs
    -- |f ('\n' : '\n' : cs) = ("\n<br>\n" ++) . f cs|
    f ('\n' : cs) = ("\n<br/>\n" ++) . f cs
    f (c@'<' : cs) = (c :) . g cs
    f (c : cs) = (c :) . f cs
    f [] = id
    g (c@'>' : cs) = (c :) . f cs
    g (c : cs) = (c :) . g cs
    g [] = id
\end{code}
%}}}

%{{{ \subsection{Laws}
\subsection{Laws}

\begin{code}
isSymEquiv :: Theorem -> Bool
isSymEquiv (("(3.2)" : _, _), _) = True
isSymEquiv _ = False

symEquivTP :: AC.TermPair Var Op Const
symEquivTP = let
    p, q :: Expr
    p = Var "p"
    q = Var "q"
  in (toTerm $ (p ≡ q ≡ q), toTerm p)
\end{code}

\begin{code}
laws :: [Theorem]
laws  = propLaws ++ substLaws ++ quantLaws ++ quantLawsP
      ++ predLaws ++ predLawsP ++ setLaws ++ relLaws
      ++ integerLaws
\end{code}

\begin{code}
laws' :: [(Theorem, AC.Term Var Op Const)]
laws' = map (\ thm@(_ , e) -> (thm , toTerm e)) laws
\end{code}

%{{{ propLaws
\begin{code}
propLaws :: [Theorem]
propLaws = let
    p, q, r, s, t :: Expr
    p = Var "p"
    q = Var "q"
    r = Var "r"
    s = Var "s"
    t = Var "t"
    u = Var "u"
  in
  [((["(1.0)"], ["Inequality"]), (t ≠ s) ≗ lnot (t ≗ s))
  ,((["(1.2)"], ["Reflexivity of ="]), t ≗ t)
  ,((["(1.3)"], ["Symmetry of ="]), (s ≗ t) ≡ (t ≗ s))
  ,((["(1.4)"], ["Transitivity of ="]), (s ≗ t) ∧ (t ≗ u) ⇒ (s ≗ u))
  ,((["(2.10)"], ["≡ equals ="]), (s ≡ t) ≡ (s ≗ t))
  ,((["(3.1)"], ["Associativity of ≡"]), ((p ≡ q) ≡ r) ≡ (p ≡ (q ≡ r)))
  ,((["(3.2)"], ["Symmetry of ≡"]), (p ≡ q) ≡ (q ≡ p))
  ,((["(3.3)"], ["Identity of ≡"]), (true ≡ q) ≡ q)
  ,((["(3.4)"], []), true)
  ,((["(3.5)"], ["Reflexivity of ≡"]), p ≡ p)
  ,((["(3.8)"], ["Definition of false"]), false ≡ lnot true)
  ,((["(3.9)"], ["Distributivity of ¬ over ≡"]), lnot (p ≡ q) ≡ (lnot p ≡ q))
  ,((["(3.10)"], ["Definition of ≢"]), (p ≢ q) ≡ lnot (p ≡ q))
  ,((["(3.11)"], []), (lnot p ≡ q) ≡ (p ≡ lnot q))
  ,((["(3.12)"], ["Double negation"]), lnot (lnot p) ≡ p)
  ,((["(3.13)"], ["Negation of false"]), lnot false ≡ true)
  ,((["(3.14)"], []), (p ≢ q) ≡ (lnot p ≡ q))
  ,((["(3.15)"], []), lnot p ≡ (p ≡ false))
  ,((["(3.16)"], ["Symmetry of ≢"]), (p ≢ q) ≡ (q ≢ p))
  ,((["(3.17)"], ["Associativity of ≢"]), ((p ≢ q) ≢ r) ≡ (p ≢ (q ≢ r)))
  ,((["(3.18)"], ["Mutual associativity of ≢ and ≡"]), ((p ≢ q) ≡ r) ≡ (p ≢ (q ≡ r)))
  ,((["(3.19)"], ["Mutual interchangeability of ≢ and ≡"]), (p ≢ (q ≡ r)) ≡ (p ≡ (q ≢ r)))
  ,((["(3.24)"], ["Symmetry of ∨"]), p ∨ q ≡ q ∨ p)
  ,((["(3.25)"], ["Associativity of ∨"]), ((p ∨ q) ∨ r) ≡ (p ∨ (q ∨ r)))
  ,((["(3.26)"], ["Idempotency of ∨"]), p ∨ p ≡ p)
  ,((["(3.27)"], ["Distributivity of ∨ over ≡"]), p ∨ (q ≡ r) ≡ p ∨ q ≡ p ∨ r)
  ,((["(3.28)"], ["Excluded Middle"]), p ∨ lnot p)
  ,((["(3.29)"], ["Zero of ∨"]), p ∨ true)
  ,((["(3.30)"], ["Identity of ∨"]), p ∨ false ≡ p)
  ,((["(3.31)"], ["Distributivity of ∨ over ∨"]), p ∨ (q ∨ r) ≡ (p ∨ q) ∨ (p ∨ r))
  ,((["(3.32)"], []), p ∨ q ≡ p ∨ lnot q ≡ p)
  ,((["(3.35)"], ["Golden Rule"]), p ∧ q ≡ (p ≡ q) ≡ p ∨ q )
  ,((["(3.36)"], ["Symmetry of ∧"]), p ∧ q ≡ q ∧ p)
  ,((["(3.37)"], ["Associativity of ∧"]), (p ∧ q) ∧ r ≡ p ∧ (q ∧ r))
  ,((["(3.38)"], ["Idempotency of ∧"]), p ∧ p ≡ p)
  ,((["(3.39)"], ["Identity of ∧"]), p ∧ true ≡ p)
  ,((["(3.40)"], ["Zero of ∧"]), p ∧ false ≡ false)
  ,((["(3.41)"], ["Distributivity of ∧ over ∧"]), p ∧ (q ∧ r) ≡ (p ∧ q) ∧ (p ∧ r))
  ,((["(3.42)"], ["Contradiction"]), p ∧ lnot p ≡ false)
  ,((["(3.43)", "(3.43a)"], ["Absorption"]), p ∧ (p ∨ q) ≡ p)
  ,((["(3.43)", "(3.43b)"], ["Absorption"]), p ∨ (p ∧ q) ≡ p)
  ,((["(3.44)", "(3.44a)"], ["Absorption"]), p ∧ (lnot p ∨ q) ≡ p ∧ q)
  ,((["(3.44)", "(3.44a)", "(3.44aN)"], ["Absorption"]), lnot p ∧ (p ∨ q) ≡ lnot p ∧ q)
  ,((["(3.44)", "(3.44b)"], ["Absorption"]), p ∨ (lnot p ∧ q) ≡ p ∨ q)
  ,((["(3.44)", "(3.44b)", "(3.44bN)"], ["Absorption"]), lnot p ∨ (p ∧ q) ≡ lnot p ∨ q)
  ,((["(3.45)"], ["Distributivity of ∨ over ∧"]), p ∨ (q ∧ r) ≡ (p ∨ q) ∧ (p ∨ r))
  ,((["(3.46)"], ["Distributivity of ∧ over ∨"]), p ∧ (q ∨ r) ≡ (p ∧ q) ∨ (p ∧ r))
  ,((["(3.47)", "(3.47a)"], ["De Morgan"]), lnot (p ∧ q) ≡ lnot p ∨ lnot q)
  ,((["(3.47)", "(3.47a)", "(3.47aN)"], ["De Morgan"]), lnot (lnot p ∧ lnot q) ≡ p ∨ q)
  ,((["(3.47)", "(3.47b)"], ["De Morgan"]), lnot (p ∨ q) ≡ lnot p ∧ lnot q)
  ,((["(3.47)", "(3.47b)", "(3.47bN)"], ["De Morgan"]), lnot (lnot p ∨ lnot q) ≡ p ∧ q)
  ,((["(3.48)"], []), p ∧ q ≡ p ∧ lnot q ≡ lnot p)
  ,((["(3.49)"], []), p ∧ (q ≡ r) ≡ p ∧ q ≡ p ∧ r ≡ p)
  ,((["(3.50)"], []), p ∧ (q ≡ p) ≡ p ∧ q)
  ,((["(3.50.1)"], []), lnot p ∧ (p ≡ q) ≡ lnot (p ∨ q))
  ,((["(3.51)"], ["Replacement"]), (p ≡ q) ∧ (r ≡ p) ≡ (p ≡ q) ∧ (r ≡ q))
  ,((["(3.52)"], ["Definition of ≡"]), (p ≡ q) ≡ (p ∧ q) ∨ (lnot p ∧ lnot q))
  ,((["(3.53)"], ["Definition of ≢"]), (p ≢ q) ≡ (lnot p ∧ q) ∨ (p ∧ lnot q))
  ,((["(3.57)"], ["Definition of ⇒"]), p ⇒ q ≡ p ∨ q ≡ q)
  ,((["(3.58)"], ["Definition of ⇐"]), p ⇐ q ≡ q ⇒ p)
  ,((["(3.59)"], ["Definition of ⇒"]), p ⇒ q ≡ lnot p ∨ q)
  ,((["(3.59)","(3.59N)"], ["Definition of ⇒"]), lnot p ⇒ q ≡ p ∨ q)
  ,((["(3.60)"], ["Definition of ⇒"]), p ⇒ q ≡ p ∧ q ≡ p)
  ,((["(3.61)"], ["Contrapositive"]), p ⇒ q ≡ lnot q ⇒ lnot p)
  ,((["(3.62)"],[]), p ⇒ (q ≡ r) ≡ p ∧ q ≡ p ∧ r)
  ,((["(3.63)"], ["Distributivity of ⇒ over ≡"]), p ⇒ (q ≡ r) ≡ p ⇒ q ≡ p ⇒ r)
  ,((["(3.64)"], ["Self-distributivity of ⇒"]), p ⇒ (q ⇒ r) ≡ (p ⇒ q) ⇒ (p ⇒ r))
  ,((["(3.65)"], ["Shunting"]), p ∧ q ⇒ r ≡ p ⇒ (q ⇒ r))
  ,((["(3.66)"], []), p ∧ (p ⇒ q) ≡ p ∧ q)
  ,((["(3.67)"], []), p ∧ (q ⇒ p) ≡ p)
  ,((["(3.68)"], []), p ∨ (p ⇒ q) ≡ true)
  ,((["(3.69)"], []), p ∨ (q ⇒ p) ≡ q ⇒ p)
  ,((["(3.70)"], []), (p ∨ q) ⇒ (p ∧ q) ≡ p ≡ q)
  ,((["(3.71)"], ["Reflexivity of ⇒"]), p ⇒ p)
  ,((["(3.72)"], ["Right zero of ⇒"]), p ⇒ true)
  ,((["(3.73)"], ["Left identity of ⇒"]), true ⇒ p ≡ p)
  ,((["(3.74)"], []), p ⇒ false ≡ lnot p)
  ,((["(3.75)"], []), false ⇒ p)
  ,((["(3.75)"], []), false ⇒ p)
  ,((["(3.76)", "(3.76a)"], ["Weakening/strengthening"]), p ⇒ p ∨ q)
  ,((["(3.76)", "(3.76b)"], ["Weakening/strengthening"]), p ∧ q ⇒ p)
  ,((["(3.76)", "(3.76c)"], ["Weakening/strengthening"]), p ∧ q ⇒ p ∨ q)
  ,((["(3.76)", "(3.76d)"], ["Weakening/strengthening"]), p ∨ (q ∧ r) ⇒ p ∨ q)
  ,((["(3.76)", "(3.76e)"], ["Weakening/strengthening"]), p ∧ q ⇒ p ∧ (q ∨ r))
  ,((["(3.76.1)"], ["Triple equivalence weakening"]), (p ≡ q ≡ r) ⇒ (p ∨ q ∨ r))
  ,((["(3.77)"], ["Modus ponens"]), p ∧ (p ⇒ q) ⇒ q)
  ,((["(3.78)"], ["Case analysis"]), (p ⇒ r) ∧ (q ⇒ r) ≡ (p ∨ q ⇒ r))
  ,((["(3.79)"], ["Case analysis"]), (p ⇒ r) ∧ (lnot p ⇒ r) ≡ r)
  ,((["(3.80)"], ["Mutual implication"]), (p ⇒ q) ∧ (q ⇒ p) ≡ (p ≡ q))
  ,((["(3.80b)"], ["Reflexivity of ⇒ wrt. ≡"]), (p ≡ q) ⇒ (p ⇒ q))
  ,((["(3.81)"], ["Antisymmetry of ⇒"]), (p ⇒ q) ∧ (q ⇒ p) ⇒ (p ≡ q))
  ,((["(3.82)", "(3.82a)"], ["Transitivity of ⇒"]), (p ⇒ q) ∧ (q ⇒ r) ⇒ (p ⇒ r))
  ,((["(3.82)", "(3.82b)"], ["Transitivity of ⇒"]), (p ≡ q) ∧ (q ⇒ r) ⇒ (p ⇒ r))
  ,((["(3.82)", "(3.82c)"], ["Transitivity of ⇒"]), (p ⇒ q) ∧ (q ≡ r) ⇒ (p ⇒ r))
  ,((["(3.90.1)"], ["(Ex.~3.45)"]), p ⇒ q ≡ lnot p ∨ lnot q ≡ lnot p)
  ,((["(3.90.2)"], ["(Ex.~3.46)"]), p ⇒ q ≡ lnot p ∧ lnot q ≡ lnot q)
  ,((["(3.91.1)", "(6.3.D)"], []), lnot p ∨ (p ⇒ q) ≡ p ⇒ q)
  ,((["(3.91.2)", "(6.3.E)"], []), p ∨ (lnot p ⇒ q) ≡  lnot p ⇒ q)
  ,((["(3.92.1)", "(6.3.A)"], []), p ⇒ (q ∧ r) ≡ (p ⇒ q) ∧ (p ⇒ r))
  ,((["(3.93.1)", "(6.3.B)"], []), (p ∧ q) ∨ (lnot p ∧ r) ≡ (p ⇒ q) ∧ (lnot p ⇒ r))
  ,((["(3.93.2)", "(6.3.C)"], []), (p ∧ q) ∨ (lnot p ∧ r) ≡ (lnot p ∨ q) ∧ (p ∨ r))
  ,((["(3.94.1)"], []), (p ∨ q) ∧ (p ∨ r) ∧ (lnot q ∨ r) ≡ (p ∨ q) ∧ (lnot q ∨ r))
  ,((["(4.1)"], []), p ⇒ (q ⇒ p))
  ,((["(4.2)"], ["Monotonicity of ∨"]), (p ⇒ q) ⇒ ((p ∨ r) ⇒ (q ∨ r)))
  ,((["(4.3)"], ["Monotonicity of ∧"]), (p ⇒ q) ⇒ ((p ∧ r) ⇒ (q ∧ r)))
  ]
\end{code}
%}}}

%{{{ substLaws
\begin{code}
substLaws :: [Theorem]
substLaws = let
    e, f, ee, p, q :: Expr
    p = Var "p"
    q = Var "q"
    e = Var "e"
    f = Var "f"
    ee = Var "E"
    sez x = Subst ee ["z"] [x]
  in
  [((["(3.83)"], ["Leibniz Axiom"]), (e ≗ f) ⇒ (sez e ≗ sez f))
  ,((["(3.84a)","(3.84)"], []), ((e ≗ f) ∧ sez e) ≡ ((e ≗ f) ∧ sez f))
  ,((["(3.84b)","(3.84)"], []), ((e ≗ f) ⇒ sez e) ≡ ((e ≗ f) ⇒ sez f))
  ,((["(3.84c)","(3.84)"], []), (q ∧ (e ≗ f) ⇒ sez e) ≡ (q ∧ (e ≗ f) ⇒ sez f))
  ,((["(3.85a)","(3.85)"], ["Replace by Ⓣ"]), (p ⇒ sez p) ≡ (p ⇒ sez true))
  ,((["(3.85b)","(3.85)"], ["Replace by Ⓣ"]), (q ∧ p ⇒ sez p) ≡ (q ∧ p ⇒ sez true))
  ,((["(3.86a)","(3.86)"], ["Replace by Ⓕ"]), (sez p ⇒ p) ≡ (sez false ⇒ p))
  ,((["(3.86b)","(3.86)"], ["Replace by Ⓕ"]), (sez p ⇒ p ∨ q) ≡ (sez false ⇒ p ∨ q))
  ,((["(3.87)"], ["Replace by Ⓣ"]), (p ∧ sez p) ≡ (p ∧ sez true))
  ,((["(3.88)"], ["Replace by Ⓕ"]), (p ∨ sez p) ≡ (p ∨ sez false))
  ,((["(3.89)"], ["Shannon"]), sez p ≡ (p ∧ sez true) ∨ (lnot p ∧ sez false))
  ]
\end{code}
%}}}

%{{{ quantLaws
\begin{code}
quantLaws :: [Theorem]
quantLaws = let
    p, q, r {- , u, f, ee -} :: Expr
    p = Var "P"
    q = Var "Q"
    r = Var "R"
    -- |u = Const '∤'|     -- identity of ⋆
    -- |f = Var "f"|
    -- |sx e1 e2 = Subst e1 ["x"] [e2]|
  in
  [{- ((["(8.11)"], ["Substitution in Quantification"]),
        (Quant '⋆' [("y", Nothing)] r p `sx` f) ≗ Quant '⋆' [("y", Nothing)] (r `sx` f) (p `sx` f))
   -}
   ((["(8.15)"], ["Quantification Distributivity"]),
       ((Quant '⋆' [("y", Nothing)] (Just r) p) ⋆
        (Quant '⋆' [("y", Nothing)] (Just r) q)) ≗
       (Quant '⋆' [("y", Nothing)] (Just r) (p ⋆ q)))
  ,((["(8.18)"], ["Range Split for idempotent ⋆"]),
       (Quant '⋆' [("y", Nothing)] (Just $ q ∨ r) p)
        ≗
       ((Quant '⋆' [("y", Nothing)] (Just q) p) ⋆
        (Quant '⋆' [("y", Nothing)] (Just r) p)))
  -- ,((["(8.13)"], ["Empty Range"]), Quant '⋆' [("y", Nothing)] (Just false) p ≗ u)
  ,((["(8.20.1)"], ["Dummy list permutation"]),
       (Quant '⋆' [("x", Nothing), ("y", Nothing)] (Just r) p)
        ≗
       (Quant '⋆' [("y", Nothing), ("x", Nothing)] (Just r) p))
  ,((["(8.20.1)","(8.20.1∧)"], ["Dummy list permutation"]),
       (Quant '∀' [("x", Nothing), ("y", Nothing)] (Just r) p)
        ≗
       (Quant '∀' [("y", Nothing), ("x", Nothing)] (Just r) p))
  ,((["(8.20.1)","(8.20.1∨)"], ["Dummy list permutation"]),
       (Quant '∃' [("x", Nothing), ("y", Nothing)] (Just r) p)
        ≗
       (Quant '∃' [("y", Nothing), ("x", Nothing)] (Just r) p))
  ]
\end{code}
%}}}

%{{{ quantLawsP
\begin{code}
quantLawsP :: [Theorem]
quantLawsP = let
    p, r, s, f, y, ee :: Expr
    p = Var "P"
    r = Var "R"
    s = Var "S"
    t = Var "T"
    -- |u = Const '∤'|     -- identity of ⋆
    f = Var "f"
    x = Var "x"
    y = Var "y"
    ee = Var "E"
    sx e1 e2 = Subst e1 ["x"] [e2]
    sz e1 e2 = Subst e1 ["z"] [e2]
  in
  [((["(8.11)"], ["Substitution in Quantification"]),
        ((Quant '⋆' [("y", Nothing)] (Just r) p) `sx` f)
        ≗ Quant '⋆' [("y", Nothing)] (Just (r `sx` f)) (p `sx` f))
  ,((["(8.14)"], ["One-point rule"]),
       (Quant '⋆' [("x", Nothing)] (Just (Var "x" ≗ ee)) p) ≗ sx p ee)
  ,((["(8.14)"], ["One-point rule"]),
       (Quant '⋆' [("x", Just t)] (Just (Var "x" ≗ ee)) p) ≗ sx p ee) -- proviso!
  ,((["(8.14)","(8.14∧)"], ["One-point rule"]),
       (Quant '∀' [("x", Nothing)] (Just (Var "x" ≗ ee)) p) ≗ sx p ee)
  ,((["(8.14)","(8.14∨)"], ["One-point rule"]),
       (Quant '∃' [("x", Nothing)] (Just (Var "x" ≗ ee)) p) ≗ sx p ee)
  ,((["(8.14)","(8.14+)"], ["One-point rule"]),
       (Quant '∑' [("x", Nothing)] (Just (Var "x" ≗ ee)) p) ≗ sx p ee)
  ,((["(8.14)","(8.14·)"], ["One-point rule"]),
       (Quant '∏' [("x", Nothing)] (Just (Var "x" ≗ ee)) p) ≗ sx p ee)
  ,((["(8.14)","(8.14∩)"], ["One-point rule"]),
       (Quant '⋂' [("x", Nothing)] (Just (Var "x" ≗ ee)) p) ≗ sx p ee)
  ,((["(8.14)","(8.14∪)"], ["One-point rule"]),
       (Quant '⋃' [("x", Nothing)] (Just (Var "x" ≗ ee)) p) ≗ sx p ee)
  ,((["(8.14)","(8.14↓)"], ["One-point rule"]),
       (Quant '⇊' [("x", Nothing)] (Just (Var "x" ≗ ee)) p) ≗ sx p ee)
  ,((["(8.14)","(8.14↑)"], ["One-point rule"]),
       (Quant '⇈' [("x", Nothing)] (Just (Var "x" ≗ ee)) p) ≗ sx p ee)
  ,((["(8.14)","(8.14⊍)"], ["One-point rule"]),
       (Quant '⨃' [("x", Nothing)] (Just (Var "x" ≗ ee)) p) ≗ sx p ee)
  ,((["(8.14)","(8.14≡)"], ["One-point rule"]),
       (Quant '≣' [("x", Nothing)] (Just (Var "x" ≗ ee)) p) ≗ sx p ee)
 -- if each quantification is defined:
  ,((["(8.19)"], ["Interchange of dummies"]),
       (Quant '⋆' [("x", Nothing)] (Just r)
        (Quant '⋆' [("y", Nothing)] (Just s) p))
        ≗
       (Quant '⋆' [("y", Nothing)] (Just s)
        (Quant '⋆' [("x", Nothing)] (Just r) p)))
  ,((["(8.19)","(8.19∧)"], ["Interchange of dummies for ∀"]),
       (Quant '∀' [("x", Nothing)] (Just r)
        (Quant '∀' [("y", Nothing)] (Just s) p))
        ≗
       (Quant '∀' [("y", Nothing)] (Just s)
        (Quant '∀' [("x", Nothing)] (Just r) p)))
  ,((["(8.19)","(8.19∨)"], ["Interchange of dummies for ∃"]),
       (Quant '∃' [("x", Nothing)] (Just r)
        (Quant '∃' [("y", Nothing)] (Just s) p))
        ≗
       (Quant '∃' [("y", Nothing)] (Just s)
        (Quant '∃' [("x", Nothing)] (Just r) p)))
  ,((["(8.20)"], ["Quantification nesting"]),
       (Quant '⋆' [("x", Nothing), ("y", Nothing)] (Just $ r ∧ s) p)
        ≗
       (Quant '⋆' [("x", Nothing)] (Just r)
        (Quant '⋆' [("y", Nothing)] (Just s) p)))
  ,((["(8.20)","(8.20∧)"], ["Quantification nesting"]),
       (Quant '∀' [("x", Nothing), ("y", Nothing)] (Just $ r ∧ s) p)
        ≗
       (Quant '∀' [("x", Nothing)] (Just r)
        (Quant '∀' [("y", Nothing)] (Just s) p)))
  ,((["(8.20)","(8.20∨)"], ["Quantification nesting"]),
       (Quant '∃' [("x", Nothing), ("y", Nothing)] (Just $ r ∧ s) p)
        ≗
       (Quant '∃' [("x", Nothing)] (Just r)
        (Quant '∃' [("y", Nothing)] (Just s) p)))
  ,((["(8.20)", "(8.20m)"], ["Quantification nesting"]),
       (Quant '⋆' [("x", Nothing), ("y", Nothing)] (Just r) p)
        ≗
       (Quant '⋆' [("x", Nothing)] Nothing
        (Quant '⋆' [("y", Nothing)] (Just r) p)))
  ,((["(8.20)","(8.20m∧)"], ["Quantification nesting"]),
       (Quant '∀' [("x", Nothing), ("y", Nothing)] (Just r) p)
        ≗
       (Quant '∀' [("x", Nothing)] Nothing
        (Quant '∀' [("y", Nothing)] (Just r) p)))
  ,((["(8.20)","(8.20m∨)"], ["Quantification nesting"]),
       (Quant '∃' [("x", Nothing), ("y", Nothing)] (Just r) p)
        ≗
       (Quant '∃' [("x", Nothing)] Nothing
        (Quant '∃' [("y", Nothing)] (Just r) p)))
  ,((["(8.20)", "(8.20n)"], ["Quantification nesting"]),
       (Quant '⋆' [("x", Nothing), ("y", Nothing)] (Just r) p)
        ≗
       (Quant '⋆' [("x", Nothing)] (Just r)
        (Quant '⋆' [("y", Nothing)] Nothing p)))
  ,((["(8.20)","(8.20n∧)"], ["Quantification nesting"]),
       (Quant '∀' [("x", Nothing), ("y", Nothing)] (Just r) p)
        ≗
       (Quant '∀' [("x", Nothing)] (Just r)
        (Quant '∀' [("y", Nothing)] Nothing p)))
  ,((["(8.20)","(8.20n∨)"], ["Quantification nesting"]),
       (Quant '∃' [("x", Nothing), ("y", Nothing)] (Just r) p)
        ≗
       (Quant '∃' [("x", Nothing)] (Just r)
        (Quant '∃' [("y", Nothing)] Nothing p)))
  ,((["(8.21)"], ["Dummy renaming"]),
       (Quant '⋆' [("x", Nothing)] (Just r) p)
        ≗
       (Quant '⋆' [("y", Nothing)] (Just $ sx r y) (sx p y)))
  ,((["(8.21)","(8.21∧)"], ["Dummy renaming"]),
       (Quant '∀' [("x", Nothing)] (Just r) p)
        ≗
       (Quant '∀' [("y", Nothing)] (Just $ sx r y) (sx p y)))
  ,((["(8.21)","(8.21∨)"], ["Dummy renaming"]),
       (Quant '∃' [("x", Nothing)] (Just r) p)
        ≗
       (Quant '∃' [("y", Nothing)] (Just $ sx r y) (sx p y)))
  ,((["(8.21p)"], ["Pair dummy expansion"]),
       (Quant '⋆' [("z", Nothing)] (Just r) p)
        ≗
       (Quant '⋆' [("x", Nothing),("y", Nothing)] (Just $ sz r (pair x y)) (sz p (pair x y))))
  ,((["(8.21p)","(8.21p∧)"], ["Pair dummy expansion"]),
       (Quant '∀' [("z", Nothing)] (Just r) p)
        ≗
       (Quant '∀' [("x", Nothing),("y", Nothing)] (Just $ sz r (pair x y)) (sz p (pair x y))))
  ,((["(8.21p)","(8.21p∨)"], ["Pair dummy expansion"]),
       (Quant '∃' [("z", Nothing)] (Just r) p)
        ≗
       (Quant '∃' [("x", Nothing),("y", Nothing)] (Just $ sz r (pair x y)) (sz p (pair x y))))
 -- if |f| has an inverse:
  ,((["(8.22)"], ["Change of dummy"]),
       (Quant '⋆' [("x", Nothing)] (Just r) p)
        ≗
       (Quant '⋆' [("y", Nothing)] (Just $ sx r $ Bin '.' f y) (sx p $ Bin '.' f y)))
  ]
\end{code}
%}}}

%{{{ predLaws
\begin{code}
predLaws :: [Theorem]
predLaws = let
    p, q, r, ee :: Expr
    p = Var "P"
    q = Var "Q"
    r = Var "R"
    ee = Var "E"
    -- |f = Var "f"|
    -- |sx e1 e2 = Subst e1 ["x"] [e2]|
  in
  [((["(8.13)", "(8.13∧)"], ["Empty Range"]), Quant '∀' [("y", Nothing)] (Just false) p ≡ true)
  ,((["(8.13)", "(8.13∨)"], ["Empty Range"]), Quant '∃' [("y", Nothing)] (Just false) p ≡ false)

  ,((["(8.13)", "(8.13≡)"], ["Empty Range"]), Quant '≣' [("y", Nothing)] (Just false) p ≡ true)
  ,((["(8.13)", "(8.13·)"], ["Empty Range"]), Quant '∏' [("y", Nothing)] (Just false) p ≗ IntLit 1)
  ,((["(8.13)", "(8.13+)"], ["Empty Range"]), Quant '∑' [("y", Nothing)] (Just false) p ≗ IntLit 0)
  ,((["(8.13)", "(8.13∩)"], ["Empty Range"]), Quant '⋂' [("y", Nothing)] (Just false) p ≗ universe)
  ,((["(8.13)", "(8.13∪)"], ["Empty Range"]), Quant '⋃' [("y", Nothing)] (Just false) p ≗ emptyset)
  ,((["(8.13)", "(8.13⊍)"], ["Empty Range"]), Quant '⨃' [("y", Nothing)] (Just false) p ≗ emptyset)
  ,((["(8.15)", "(8.15∧)"], ["Distributivity of ∀ over ∧"]),
       ((Quant '∀' [("y", Nothing)] (Just r) p) ∧
        (Quant '∀' [("y", Nothing)] (Just r) q)) ≡
       (Quant '∀' [("y", Nothing)] (Just r) (p ∧ q)))
  ,((["(8.15)", "(8.15∨)"], ["Distributivity of ∃ over ∨"]),
       ((Quant '∃' [("y", Nothing)] (Just r) p) ∨
        (Quant '∃' [("y", Nothing)] (Just r) q)) ≡
       (Quant '∃' [("y", Nothing)] (Just r) (p ∨ q)))
  ,((["(8.18)", "(8.18∧)"], ["Range Split for ∀"]),
       (Quant '∀' [("y", Nothing)] (Just $ q ∨ r) p)
        ≡
       ((Quant '∀' [("y", Nothing)] (Just q) p) ∧
        (Quant '∀' [("y", Nothing)] (Just r) p)))
  ,((["(8.18)", "(8.18∨)"], ["Range Split for ∃"]),
       (Quant '∃' [("y", Nothing)] (Just $ q ∨ r) p)
        ≡
       ((Quant '∃' [("y", Nothing)] (Just q) p) ∨
        (Quant '∃' [("y", Nothing)] (Just r) p)))
  ,((["(9.2)"], ["Trading for ∀"]),
        Quant '∀' [("x", Nothing)] (Just r) p ≡
        Quant '∀' [("x", Nothing)] Nothing (r ⇒ p) )
  ,((["(9.3a)", "(9.3)"], ["Trading for ∀"]),
        Quant '∀' [("x", Nothing)] (Just r) p ≡
        Quant '∀' [("x", Nothing)] Nothing (lnot r ∨ p) )
  ,((["(9.3b)", "(9.3)"], ["Trading for ∀"]),
        Quant '∀' [("x", Nothing)] (Just r) p ≡
        Quant '∀' [("x", Nothing)] Nothing (r ∧ p ≡ r) )
  ,((["(9.3c)", "(9.3)"], ["Trading for ∀"]),
        Quant '∀' [("x", Nothing)] (Just r) p ≡
        Quant '∀' [("x", Nothing)] Nothing (r ∨ p ≡ p) )
  ,((["(9.4a)", "(9.4)"], ["Trading for ∀"]),
        Quant '∀' [("x", Nothing)] (Just $ q ∧ r) p ≡
        Quant '∀' [("x", Nothing)] (Just q) (r ⇒ p) )
  ,((["(9.4b)", "(9.4)"], ["Trading for ∀"]),
        Quant '∀' [("x", Nothing)] (Just $ q ∧ r) p ≡
        Quant '∀' [("x", Nothing)] (Just q) (lnot r ∨ p) )
  ,((["(9.4c)", "(9.4)"], ["Trading for ∀"]),
        Quant '∀' [("x", Nothing)] (Just $ q ∧ r) p ≡
        Quant '∀' [("x", Nothing)] (Just q) (r ∧ p ≡ r) )
  ,((["(9.4d)", "(9.4)"], ["Trading for ∀"]),
        Quant '∀' [("x", Nothing)] (Just $ q ∧ r) p ≡
        Quant '∀' [("x", Nothing)] (Just q) (r ∨ p ≡ p) )
  ,((["(9.8)"], []), Quant '∀' [("x", Nothing)] (Just r) true ≡ true)
  ,((["(9.9)"], []),
       Quant '∀' [("x", Nothing)] (Just r) (p ≡ q) ⇒
       (Quant '∀' [("x", Nothing)] (Just r) p ≡
        Quant '∀' [("x", Nothing)] (Just r) q))
  ,((["(9.10)"], ["Range strengthening for ∀"]),
       Quant '∀' [("x", Nothing)] (Just $ q ∨ r) p ⇒
       Quant '∀' [("x", Nothing)] (Just q) p)
  ,((["(9.11)"], ["Body weakening for ∀"]),
       Quant '∀' [("x", Nothing)] (Just r) (p ∧ q) ⇒
       Quant '∀' [("x", Nothing)] (Just r) p)
  ,((["(9.12)"], ["Monotonicity of ∀"]),
       Quant '∀' [("x", Nothing)] (Just r) (p ⇒ q) ⇒
       (Quant '∀' [("x", Nothing)] (Just r) p ⇒
        Quant '∀' [("x", Nothing)] (Just r) q))
  ,((["(9.12a)","(9.12)"], ["Range-Antitonicity of ∀"]),
       Quant '∀' [("x", Nothing)] Nothing (q ⇒ r) ⇒
       (Quant '∀' [("x", Nothing)] (Just r) p ⇒
        Quant '∀' [("x", Nothing)] (Just q) p))
  ,((["(9.13)"], ["Instantiation"]),
       Quant '∀' [("x", Nothing)] Nothing p ⇒ Subst p ["x"] [ee])
  ,((["(9.17)"], ["Axiom, Generalised DeMorgan"]),
       Quant '∃' [("x", Nothing)] (Just r) p ≡
       lnot (Quant '∀' [("x", Nothing)] (Just r) (lnot p)))
  ,((["(9.18a)","(9.18)"], ["Generalised DeMorgan"]),
       lnot (Quant '∃' [("x", Nothing)] (Just r) (lnot p)) ≡
       Quant '∀' [("x", Nothing)] (Just r) p)
  ,((["(9.18b)","(9.18)"], ["Generalised DeMorgan"]),
       lnot (Quant '∃' [("x", Nothing)] (Just r) p) ≡
       Quant '∀' [("x", Nothing)] (Just r) (lnot p))
  ,((["(9.18c)","(9.18)"], ["Generalised DeMorgan"]),
       Quant '∃' [("x", Nothing)] (Just r)  (lnot p) ≡
       lnot (Quant '∀' [("x", Nothing)] (Just r) p))
  ,((["(9.19)"], ["Trading for ∃"]),
       Quant '∃' [("x", Nothing)] (Just r) p ≡
       Quant '∃' [("x", Nothing)] Nothing (r ∧ p))
  ,((["(9.19)"], ["Trading for ∃"]),
       Quant '∃' [("x", Nothing)] (Just r) p ≡
       Quant '∃' [("x", Nothing)] (Just p) r)
  ,((["(9.20)"], ["Trading for ∃"]),
       Quant '∃' [("x", Nothing)] (Just $ q ∧ r) p ≡
       Quant '∃' [("x", Nothing)] (Just q) (r ∧ p))
  ,((["(9.21)"], ["Distributivity of ∧ over ∃"]),
       p ∧ Quant '∃' [("x", Nothing)] (Just r) q ≡
       Quant '∃' [("x", Nothing)] (Just r) (p ∧ q))
  ,((["(9.24)"], []),
       Quant '∃' [("x", Nothing)] (Just r) false ≡ false)
  ,((["(9.25)"], ["Range weakening for ∃"]),
       Quant '∃' [("x", Nothing)] (Just r) p ⇒
       Quant '∃' [("x", Nothing)] (Just $ q ∨ r) p)
  ,((["(9.26)"], ["Body weakening for ∃"]),
       Quant '∃' [("x", Nothing)] (Just r) p ⇒
       Quant '∃' [("x", Nothing)] (Just r) (p ∨ q))
  ,((["(9.27)"], ["Monotonicity of ∃"]),
       Quant '∀' [("x", Nothing)] (Just r) (p ⇒ q) ⇒
       (Quant '∃' [("x", Nothing)] (Just r) p ⇒
        Quant '∃' [("x", Nothing)] (Just r) q))
  ,((["(9.27a)","(9.27)"], ["Range-Monotonicity of ∃"]),
       Quant '∀' [("x", Nothing)] Nothing (q ⇒ r) ⇒
       (Quant '∃' [("x", Nothing)] (Just q) p ⇒
        Quant '∃' [("x", Nothing)] (Just r) p))
  ,((["(9.28)"], ["∃-Introduction"]),
       Subst p ["x"] [ee] ⇒
       Quant '∃' [("x", Nothing)] Nothing p)
  ]
\end{code}
%}}}

%{{{ predLawsP
\begin{code}
predLawsP :: [Theorem]
predLawsP = let
    p, q, r :: Expr
    p = Var "P"
    q = Var "Q"
    r = Var "R"
    s = Var "S"
    -- |f = Var "f"|
    -- |sx e1 e2 = Subst e1 ["x"] [e2]|
  in
  [((["(8.19)","(8.19∧)"], ["Interchange of dummies in ∀"]),
       (Quant '∀' [("x", Nothing)] (Just r)
        (Quant '∀' [("y", Nothing)] (Just s) p))
       ≡
       (Quant '∀' [("y", Nothing)] (Just s)
        (Quant '∀' [("x", Nothing)] (Just r) p)))
  ,((["(8.19)","(8.19∨)"], ["Interchange of dummies in ∃"]),
       (Quant '∃' [("x", Nothing)] (Just r)
        (Quant '∃' [("y", Nothing)] (Just s) p))
       ≡
       (Quant '∃' [("y", Nothing)] (Just s)
        (Quant '∃' [("x", Nothing)] (Just r) p)))
  ,((["(9.5)"], ["Distributivity of ∨ over ∀"]),
       p ∨ (Quant '∀' [("x", Nothing)] (Just r) q)
       ≡
       (Quant '∀' [("x", Nothing)] (Just r) (p ∨ q)))
  ,((["(9.6)"], []),
       (Quant '∀' [("x", Nothing)] (Just r) p)
       ≡
       p ∨ (Quant '∀' [("x", Nothing)] Nothing (lnot r)))
  ,((["(9.7)"], ["Distributivity of ∧ over ∀"]),
       lnot (Quant '∀' [("x", Nothing)] Nothing (lnot r))
       ⇒
       ((Quant '∀' [("x", Nothing)] (Just r) (p ∧ q))
        ≡
        p ∧ (Quant '∀' [("x", Nothing)] (Just r) q)))
  ,((["(9.7v)","(9.7)"], ["Distributivity of ∧ over ∀"]),
       (Quant '∃' [("x", Nothing)] Nothing r)
       ⇒
       (p ∧ (Quant '∀' [("x", Nothing)] (Just r) q)
        ≡
        (Quant '∀' [("x", Nothing)] (Just r) (p ∧ q))))
  ,((["(9.21)"], ["Distributivity of ∧ over ∃"]),
       p ∧ (Quant '∃' [("x", Nothing)] (Just r) q)
       ≡
       (Quant '∃' [("x", Nothing)] (Just r) (p ∧ q)))
  ,((["(9.22)"], []),
       (Quant '∃' [("x", Nothing)] (Just r) p)
       ≡
       p ∧ (Quant '∃' [("x", Nothing)] Nothing r))
  ,((["(9.23)"], ["Distributivity of ∨ over ∃"]),
       (Quant '∃' [("x", Nothing)] Nothing r)
       ⇒
       (p ∨ (Quant '∃' [("x", Nothing)] (Just r) q)
        ≡
        (Quant '∃' [("x", Nothing)] (Just r) (p ∨ q))))
  ,((["(9.29)"], ["Interchange of quantifications"]),
        Quant '∃' [("x", Nothing)] (Just r)
         (Quant '∀' [("y", Nothing)] (Just q) p)
        ≡
        Quant '∀' [("y", Nothing)] (Just q)
         (Quant '∃' [("x", Nothing)] (Just r) p))
  ]
\end{code}
%}}}

%{{{ setLaws
\begin{code}
setLaws :: [Theorem]
setLaws = let
    v, x, pp, qq, rr, ss, tt, uu, vv, ee :: Expr
    v = Var "v"
    x = Var "x"
    pp = Var "P"
    qq = Var "Q"
    rr = Var "R"
    ss = Var "S"
    tt = Var "T"
    uu = Var "U"
    vv = Var "V"
    ee = Var "E"
    ff = Var "F"
    inEnum es = ((["(11.2e)"], ["Membership in set enumeration"]), x ∈ SetEnum es ≡ foldr1 (∨) (map (x ≗) es))
    evarlist :: Int -> [Expr]
    evarlist k = map (Var . ('e' :) . show) [1 .. k]
  in map (inEnum . evarlist) [1 .. 20] ++ -- arbitrary limit
  [((["(11.0)"], ["Set non-membership"]), (x ∉ ss) ≗ lnot (x ∈ ss))
  ,((["(11.2f)"], ["Membership in empty set"]), lnot (x ∈ SetEnum []))
  ,((["(11.2f)"], ["Membership in empty set"]), lnot (x ∈ emptyset))
  ,((["(11.3)"], ["Set membership"]),
      ff ∈ SetCompr [("x", Nothing)] (Just rr) (Just ee)
      ≡
      Quant '∃' [("x", Nothing)] (Just rr) (ee ≗ ff))
  ,((["(11.4)"], ["Set extensionality"]),
      (ss ≗ tt) ≡ Quant '∀' [("x", Nothing)] Nothing (x ∈ ss ≡ x ∈ tt)) -- proviso!
  ,((["(11.4e)"], ["Empty set"]), SetCompr [("x", Nothing)] (Just false) (Just ee) ≗ emptyset)
  ,((["(11.7)"], []), x ∈ SetCompr [("x", Nothing)] (Just rr) Nothing ≡ rr)
  ,((["(11.9)"], []),
     SetCompr [("x", Nothing)] (Just rr) Nothing ≗
     SetCompr [("x", Nothing)] (Just rr) Nothing
      ≡  (Quant '∀' [("x", Nothing)] Nothing (qq ≡ rr)))
  ,((["(11.12)"], ["Set size"]), setSize ss ≗ (Quant '∑' [("x", Nothing)] (Just (x ∈ ss)) (IntLit 1))) -- proviso!
  ,((["(11.13)"], ["Subset"]),
      (ss ⊆ tt) ≡ Quant '∀' [("x", Nothing)] (Just (x ∈ ss)) (x ∈ tt))
  ,((["(11.14)"], ["Proper subset"]),  ss ⊂ tt  ≡  ss ⊆ tt ∧ ss ≠ tt)
  ,((["(11.15)"], ["Superset"]),  tt ⊇ ss  ≡  ss ⊆ tt)
  ,((["(11.16)"], ["Proper superset"]),  tt ⊃ ss ≡ ss ⊂ tt)
  ,((["(11.17u)"], ["Universe"]),  v ∈ universe)
  ,((["(11.17)"], ["Set complement"]),  v ∈ setCompl ss ≡ v ∉ ss)
  ,((["(11.17)"], ["Set complement"]),  v ∈ setCompl ss ≡ lnot (v ∈ ss))
  ,((["(11.20)"], ["Union"]), v ∈ ss ∪ tt  ≡ v ∈ ss  ∨  v ∈ tt)
  ,((["(11.21)"], ["Intersection"]), v ∈ ss ∩ tt  ≡ v ∈ ss  ∧  v ∈ tt)
  ,((["(11.22)"], ["Set difference"]), v ∈ ss ╌ tt  ≡ v ∈ ss  ∧  v ∉ tt)
  ,((["(11.22)"], ["Set difference"]), v ∈ ss ╌ tt  ≡ v ∈ ss  ∧  lnot (v ∈ tt))
  ,((["(11.22c)"], ["Set difference via complement"]), ss ╌ tt  ≡ ss ∩ setCompl tt)
  ,((["(11.23)"], ["Power set"]), v ∈ power ss  ≡ v ⊆ ss)
  ,((["(11.26)"], ["Symmetry of ∪"]), ss ∪ tt ≗ tt ∪ ss)
  ,((["(11.27)"], ["Associativity of ∪"]), (ss ∪ tt) ∪ uu ≗ ss ∪ (tt ∪ uu))
  ,((["(11.28)"], ["Idempotency of ∪"]), ss ∪ ss ≗ ss)
  ,((["(11.30)"], ["Identity of ∪"]), ss ∪ emptyset ≗ ss)
  ,((["(11.31)"], ["Set weakening"]), ss ⊆ ss ∪ tt)
  ,((["(11.33)"], ["Symmetry of ∩"]), ss ∩ tt ≗ tt ∩ ss)
  ,((["(11.34)"], ["Associativity of ∩"]), (ss ∩ tt) ∩ uu ≗ ss ∩ (tt ∩ uu))
  ,((["(11.35)"], ["Idempotency of ∩"]), ss ∩ ss ≗ ss)
  ,((["(11.36)"], ["Zero of ∩"]), ss ∩ emptyset ≗ emptyset)
  ,((["(11.38)"], ["Set strengthening"]), ss ∩ tt ⊆ ss)
  ,((["(11.40)"], ["Distributivity of ∪ over ∩"]), ss ∪ (tt ∩ uu) ≗ (ss ∪ tt) ∩ (ss ∪ uu))
  ,((["(11.41)"], ["Distributivity of ∩ over ∪"]), ss ∩ (tt ∪ uu) ≗ (ss ∩ tt) ∪ (ss ∩ uu))
  ,((["(11.43)"], ["Monotonicity of ∪"]), ss ⊆ tt ∧ uu ⊆ vv ⇒ (ss ∪ uu) ⊆ (tt ∪ vv))
  ,((["(11.44)"], ["Monotonicity of ∩"]), ss ⊆ tt ∧ uu ⊆ vv ⇒ (ss ∩ uu) ⊆ (tt ∩ vv))
  ,((["(11.45)"], []),  ss ⊆ tt ≡ ss ∪ tt ≗ tt)
  ,((["(11.46)"], []),  ss ⊆ tt ≡ ss ∩ tt ≗ ss)
  ,((["(11.48)"], []),  ss ∩ tt ≗ emptyset ≡ Quant '∀' [("x", Nothing)] Nothing (x ∈ ss ⇒ x ∉ tt))
  ,((["(11.50)"], []),  ss ╌ tt ⊆ ss)
  ,((["(11.51)"], []),  ss ╌ emptyset ≗ ss)
  ,((["(11.52)"], []),  ss ∩ (tt ╌ ss) ≗ emptyset)
  ,((["(11.53)"], []),  ss ∪ (tt ╌ ss) ≗ ss ∪ tt)
  ,((["(11.54)"], []),  ss ╌ (tt ∪ uu) ≗ (ss ╌ tt) ∩ (ss ╌ uu))
  ,((["(11.55)"], []),  ss ╌ (tt ∩ uu) ≗ (ss ╌ tt) ∪ (ss ╌ uu))
  ,((["(11.56)"], []),
     (Quant '∀' [("x", Nothing)] Nothing (pp ⇒ qq))
     ≡      SetCompr [("x", Nothing)] (Just pp) Nothing ⊆
            SetCompr [("x", Nothing)] (Just qq) Nothing)
  ,((["(11.57)"], ["Antisymmetry of ⊆"]),  ss ⊆ tt ∧ tt ⊆ ss ≡ ss ≗ tt)
  ,((["(11.58)"], ["Reflexivity of ⊆"]),  ss ⊆ ss)
  ,((["(11.59)"], ["Transitivity of ⊆"]),  ss ⊆ tt ∧ tt ⊆ uu ≡ ss ⊆ uu)
  ,((["(11.60)"], []),  emptyset ⊆ ss)
  ,((["(11.61)"], []),  ss ⊂ tt ≡ ss ⊆ tt ∧ tt ⊈ ss)
  ,((["(11.62)"], []),  ss ⊂ tt ≡ ss ⊆ tt ∧ Quant '∃' [("x", Nothing)] (Just $ x ∈ tt) (x ∉ ss)) -- proviso!
  ,((["(11.63)"], []),  ss ⊆ tt ≡ ss ⊆ tt ∨ ss ≗ tt)
  ,((["(11.64)"], []),  ss ⊄ ss)
  ,((["(11.65)"], []),  ss ⊂ tt ⇒ ss ⊆ tt)
  ,((["(11.66)"], []),  ss ⊂ tt ⇒ tt ⊈ ss)
  ,((["(11.67)"], []),  ss ⊆ tt ⇒ tt ⊄ ss)
  ,((["(11.68)"], []),  ss ⊆ tt ∧ uu ⊈ tt ⇒ uu ⊈ ss)
  ,((["(11.70a)", "(11.70)"], ["Transitivity of ⊂"]),  ss ⊆ tt ∧ tt ⊂ uu ⇒ ss ⊂ uu)
  ,((["(11.70b)", "(11.70)"], ["Transitivity of ⊂"]),  ss ⊂ tt ∧ tt ⊆ uu ⇒ ss ⊂ uu)
  ,((["(11.70c)", "(11.70)"], ["Transitivity of ⊂"]),  ss ⊂ tt ∧ tt ⊂ uu ⇒ ss ⊂ uu)
  ,((["(11.91.1)"], ["Axiom, Symmetric Difference"]),  (ss ⊖ tt) ≗ ((ss ╌ tt) ∪ (tt ╌ ss)))
  ,((["(11.91.2)"], ["Symmetric Difference"]),   v ∈ ss ⊖ tt  ≡ v ∈ ss ≢ v ∈ tt)
  ,((["(11.91.3)"], ["Symmetry of ⊖"]),   ss ⊖ tt  ≗ tt ⊖ ss)
  ,((["(11.91.4)"], ["Associativity of ⊖"]),   (ss ⊖ tt) ⊖ uu  ≗ ss ⊖ (tt ⊖ uu))
  ,((["(11.91.5)"], ["Identity of ⊖"]),   ss ⊖ emptyset  ≗ ss)
  ,((["(11.91.6)"], ["Self-inverse of ⊖"]),   ss ⊖ ss  ≗ emptyset)
  ,((["(11.91.7)"], ["Cancellation of ⊖"]),   ss ⊖ tt ⊖ tt  ≗ ss)
  ,((["(11.91.8)"], ["Distributivity of ∩ over ⊖"]),   ss ∩ (tt ⊖ uu) ≗ (ss ∩ tt) ⊖ (ss ∩ uu))
  ]
\end{code}
%}}}

%{{{ relLaws
\begin{code}
relLaws :: [Theorem]
relLaws = let
    b, b', c, c', p, q, x, y, qq, rr, ss, tt, uu, vv :: Expr
    b   = Var "b"
    b'  = Var "b'"
    c   = Var "c"
    c'  = Var "c'"
    p   = Var "p"
    q   = Var "q"
    x   = Var "x"
    y   = Var "y"
    z   = Var "z"
    aa = Var "A"
    bb = Var "B"
    qq = Var "Q"
    rr = Var "R"
    ss = Var "S"
    tt = Var "T"
    uu = Var "U"
    vv = Var "V"
  in
  [((["(14.2)"], ["Pair equality"]),
      pair b c ≗ pair b' c'  ≡   b ≗ b' ∧ c ≗ c')
  ,((["(14.3)"], ["Axiom, Cross product"]),
      ss × tt  ≗  SetCompr [("b", Nothing),("c", Nothing)] (Just (b ∈ ss ∧ c ∈ tt)) (Just (pair b c)))
  ,((["(14.4)"], ["Cartesian product membership"]),
      pair x y ∈ ss × tt  ≡   x ∈ ss ∧ y ∈ tt)
  ,((["(14.4p)", "(14.4p.a)"], ["Pair projection"]),   proj1 @. (pair b c) ≗ b)
  ,((["(14.4p)", "(14.4p.b)"], ["Pair projection"]),   proj2 @. (pair b c) ≗ c)
  ,((["(14.2p)"], ["Pair equality"]),
      p ≗ q  ≡   proj1 @. p ≗ proj1 @. q ∧ proj2 @. p ≗ proj2 @. q)
  ,((["(14.2e)"], ["Pair extensionality"]), p ≗ pair (proj1 @. p) (proj2 @. p))
  ,((["(14.5)"], []),
      pair x y ∈ ss × tt  ≡   pair y x ∈ tt × ss)
  ,((["(14.6)"], []),             ss ≗ emptyset  ⇒  ss × tt ≗ tt × ss ≗ emptyset)
  ,((["(14.6a)", "(14.6)"], []),  ss ≗ emptyset  ⇒  ss × tt ≗ emptyset)
  ,((["(14.6b)", "(14.6)"], []),  ss ≗ emptyset  ⇒  tt × ss ≗ emptyset)
  ,((["(14.6c)", "(14.6)"], []),  ss ≗ emptyset  ⇒  ss × tt ≗ tt × ss)
  ,((["(14.7)"], []),  ss × tt ≗ tt × ss  ≡   ss ≗ emptyset ∨ ss ≗ emptyset ∨ ss ≗ tt)
  ,((["(14.8a)", "(14.8)"], ["Distributivity of × over ∪"]),  ss × (tt ∪ uu) ≗ (ss × tt) ∪ (ss × uu))
  ,((["(14.8b)", "(14.8)"], ["Distributivity of × over ∪"]),  (ss ∪ tt) × uu ≗ (ss × uu) ∪ (tt × uu))
  ,((["(14.9a)", "(14.9)"], ["Distributivity of × over ∩"]),  ss × (tt ∩ uu) ≗ (ss × tt) ∩ (ss × uu))
  ,((["(14.9b)", "(14.9)"], ["Distributivity of × over ∩"]),  (ss ∩ tt) × uu ≗ (ss × uu) ∩ (tt × uu))
  ,((["(14.10a)", "(14.10)"], ["Distributivity of × over -"]),  ss × (tt ╌ uu) ≗ (ss × tt) ╌ (ss × uu))
  ,((["(14.10b)", "(14.10)"], ["Distributivity of × over -"]),  (ss ╌ tt) × uu ≗ (ss × uu) ╌ (tt × uu))
  ,((["(14.11a)", "(14.11)"], ["Monotonicity of ×"]),  tt ⊆ uu  ⇒  ss × tt ⊆ ss × uu)
  ,((["(14.11b)", "(14.11)"], ["Monotonicity of ×"]),  tt ⊆ uu  ⇒  tt × ss ⊆ uu × ss)
  ,((["(14.12)"], ["Monotonicity of ×"]),  ss ⊆ uu ∧ tt ⊆ vv  ⇒  ss × tt ⊆ uu × vv)
  ,((["(14.13)"], []),  ss × tt ⊆ ss × uu  ∧ ss ≠ emptyset  ⇒ tt ⊆ uu)
  ,((["(14.14)"], []),  (ss ∩ tt) × (uu ∩ vv) ≗ (ss × uu) ∩ (tt × vv))
  ,((["(14.16R)"], []),  (aa ↔ bb) ≗ power (aa × bb))
  ,((["(14.16r)"], []),  rr ∈ (aa ↔ bb) ≡ rr ⊆ (aa × bb))
  ,((["(14.16I)"], []),  pair x y ∈ relId @. bb ≡ x ≗ y ∧ y ∈ bb)
  ,((["(14.16)"], ["Domain"]),
       relDom @. rr ≗
       SetCompr [("b", Nothing)]
          (Just (Quant '∃' [("c", Nothing)] Nothing (pair b c ∈ rr)))
          Nothing)
  ,((["(14.17)"], ["Range"]),
       relDom @. rr ≗
       SetCompr [("c", Nothing)]
          (Just (Quant '∃' [("b", Nothing)] Nothing (pair b c ∈ rr)))
          Nothing)
  ,((["(11.13r)"], ["Relation inclusion"]),
      (ss ⊆ tt) ≡ Quant '∀' [("x", Nothing),("y", Nothing)] (Just (pair x y ∈ ss)) (pair x y ∈ tt))
  ,((["(11.17)", "(11.17r)"], ["Relation complement"]),  pair x y ∈ setCompl ss ≡ pair x y ∉ ss)
  ,((["(11.17)", "(11.17r)"], ["Relation complement"]),  pair x y ∈ setCompl ss ≡ lnot (pair x y ∈ ss))
  ,((["(11.20)", "(11.20r)"], ["Relation union"]),  pair x y ∈ (rr ∪ ss) ≡ pair x y ∈ rr ∨ pair x y ∈ ss)
  ,((["(11.21)", "(11.21r)"], ["Relation intersection"]),  pair x y ∈ (rr ∩ ss) ≡ pair x y ∈ rr ∧ pair x y ∈ ss)
  ,((["(11.22)", "(11.22r)"], ["Relation difference"]),  pair x y ∈ (rr ╌ ss) ≡ pair x y ∈ rr ∧ lnot (pair x y ∈ ss))
  ,((["(14.18)", "(14.18r)"], ["Relation converse"]),  pair x y ∈ relConv rr ≡ pair y x ∈ rr)
  ,((["(14.19a)", "(14.19)"], ["Domain of converse"]), relDom @. relConv rr ≗ relRan @. rr)
  ,((["(14.19b)", "(14.19)"], ["Range of converse"]), relRan @. relConv rr ≗ relDom @. rr)
  ,((["(14.19c)", "(14.19)"], ["Converse in relation set"]), rr ∈ aa ↔ bb  ≡  relConv rr ∈ bb ↔ aa)
  ,((["(14.19d)", "(14.19)"], ["Converse selfinverse"]), relConv (relConv rr) ≗ rr)
  ,((["(14.19e)", "(14.19)"], ["Monotonicity of converse"]), rr ⊆ ss ≡  relConv rr ⊆ relConv ss)
  ,((["(14.20)"], ["Relation composition"]), pair x z ∈ (rr ⨾ ss) ≡ (Quant '∃' [("y", Nothing)] Nothing $ pair x y ∈ rr ∧ pair y z ∈ ss))
  ,((["(14.22)"], ["Associativity of ⨾"]), (qq ⨾ rr) ⨾ ss ≗ qq ⨾ (rr ⨾ ss))
  ,((["(14.22iL)", "(14.22i)"], ["Left-identity of ⨾"]), rr ∈ aa ↔ bb  ⇒  (relId @. aa) ⨾ rr ≗ rr)
  ,((["(14.22iR)", "(14.22i)"], ["Right-identity of ⨾"]), rr ∈ aa ↔ bb  ⇒  rr ⨾ (relId @. b) ≗ rr)
  ]
\end{code}
%}}}

%{{{ integerLaws
% i, j, k, m, n, p, q,
%     i   = Var "i"
%     j   = Var "j"
%     k   = Var "k"
%     m   = Var "m"
%     n   = Var "n"
%     p   = Var "p"
%     q   = Var "q"
\begin{code}
integerLaws :: [Theorem]
integerLaws = let
    a, b, c, d, x, y, z :: Expr
    a   = Var "a"
    b   = Var "b"
    c   = Var "c"
    d   = Var "d"
    x   = Var "x"
    y   = Var "y"
    z   = Var "z"
  in
  [((["(1.9)"], ["Cancellation of division by 2"]),
      (IntLit 2 · x) ÷ IntLit 2 ≗ x)
  ,((["(15.1)", "(15.1a)"], ["Axiom, Associativity of +", "Associativity"]),
      (a ⊹ b) ⊹ c ≗ a ⊹ (b ⊹ c))
  ,((["(15.1)", "(15.1b)"], ["Axiom, Associativity of ·", "Associativity"]),
      (a · b) · c ≗ a · (b · c))
  ,((["(15.2)", "(15.2a)"], ["Axiom, Symmetry of +", "Symmetry"]),
      a ⊹ b ≗ b ⊹ a)
  ,((["(15.2)", "(15.2b)"], ["Axiom, Symmetry of ·", "Symmetry"]),
      a · b ≗ b · a)
  ,((["(15.3)", "(15.3a)"], ["Axiom, Additive identity", "Identity of +"]),
      IntLit 0 ⊹ a ≗ a)
  ,((["(15.3)", "(15.3b)"], ["Axiom, Additive identity", "Identity of +"]),
      a ⊹ IntLit 0 ≗ a)
  ,((["(15.4)", "(15.4a)"], ["Axiom, Multiplicative identity", "Identity of ·"]),
      IntLit 1 · a ≗ a)
  ,((["(15.4)", "(15.4b)"], ["Axiom, Multiplicative identity", "Identity of ·"]),
      a · IntLit 1 ≗ a)
  ,((["(15.5)", "(15.5a)"], ["Axiom, Distributivity of · over +", "Distributivity"]),
      a · (b ⊹ c) ≗ a · b ⊹ a · c)
  ,((["(15.5)", "(15.5b)"], ["Axiom, Distributivity of · over +", "Distributivity"]),
      (b ⊹ c) · a ≗ b · a ⊹ c · a)
  ,((["(15.6)", "(15.6a)"], ["Axiom, Additive inverse"]),
      Quant '∃' [("x", Nothing)] Nothing (x ⊹ a ≗ IntLit 0))
  ,((["(15.6)", "(15.6b)"], ["Axiom, Additive inverse"]),
      Quant '∃' [("x", Nothing)] Nothing (a ⊹ x ≗ IntLit 0))
  ,((["(15.7)", "(15.7a)"], ["Axiom, Cancellation of ·", "Cancellation"]),
      (c ≠ IntLit 0) ⇒ (c · a ≗ c · b) ≡ a ≗ b)
  ,((["(15.7)", "(15.7a)"], ["Axiom, Cancellation of ·", "Cancellation"]),
      (c ≠ IntLit 0) ⇒ (a · c ≗ b · c) ≡ a ≗ b)
  ,((["(15.8)", "(15.8a)"], ["Cancellation of +", "Cancellation"]),
      (a ⊹ b ≗ a ⊹ c) ≡ b ≗ c)
  ,((["(15.8)", "(15.8b)"], ["Cancellation of +", "Cancellation"]),
      (b ⊹ a ≗ c ⊹ a) ≡ b ≗ c)
  ,((["(15.9)"], ["Multiplicative zero", "Zero", "Zero of ·"]),
      a · IntLit 0 ≗ IntLit 0)
  ,((["(15.10)", "(15.10a)"], ["Unique identity of +"]),
      (a ⊹ z ≗ a) ≡ z ≗ IntLit 0)
  ,((["(15.10)", "(15.10b)"], ["Unique identity of ·"]),
      (a ≠ IntLit 0) ⇒ (a · z ≗ a ≡ z ≗ IntLit 1))
  ,((["(15.11)"], ["No zero factors"]),
      (a · b ≗ IntLit 0) ≡ a ≗ IntLit 0 ∨ b ≗ IntLit 0)

  ,((["(15.12)"], ["Unique additive inverse"]),
      (x ⊹ a ≗ IntLit 0 ∧ y ⊹ a ≗ IntLit 0 ⇒ x ≗ y))

  ,((["(15.13)"], ["Axiom, Unary minus"]),
      (a ⊹ preMinus a ≗ IntLit 0))
  ,((["(15.14)"], ["Axiom, Subtraction"]),
      (a ╌ b ≗ a ⊹ preMinus b))

  ,((["(15.15)"], []), (x ⊹ a ≗ IntLit 0  ≡  x ≗ preMinus a))
  ,((["(15.16)"], []), (preMinus a ≗ preMinus b  ≡  a ≗ b))
  ,((["(15.17)"], []), (preMinus (preMinus a) ≗ a))
  ,((["(15.18)"], []), (preMinus (IntLit 0) ≗ IntLit 0))
  ,((["(15.19)"], []), (preMinus (a ⊹ b) ≗ preMinus a ⊹ preMinus b))
  ,((["(15.20)"], []), (preMinus a ≗ preMinus (IntLit 1) · a))
  ,((["(15.21)"], []), (preMinus a · b ≗ a · preMinus b))
  ,((["(15.22)"], []), (a · preMinus b ≗ preMinus (a · b)))
  ,((["(15.23)"], []), (preMinus a · preMinus b ≗ a · b))
  ,((["(15.24)"], []), (a ╌ IntLit 0 ≗ a))
  ,((["(15.24a)"], []), (a ╌ a ≗ IntLit 0))
  ,((["(15.25)"], []), ((a ╌ b) ⊹ (c ╌ d) ≗ (a ⊹ c) ╌ (b ⊹ d)))
  ,((["(15.25a)"], ["Mutual associativity of $+$ and $-$"]), (a ⊹ (b ╌ c) ≗ (a ⊹ b) ╌ c))
  ,((["(15.25b)"], []), ((a ╌ (b ⊹ c)) ≗ ((a ╌ b) ╌ c)))
  ,((["(15.25c)"], []), (((a ╌ b) ⊹ (b ╌ c)) ≗ (a ╌ c)))
  ,((["(15.26)"], []), ((a ╌ b) ╌ (c ╌ d) ≗ (a ⊹ d) ╌ (b ⊹ c)))
  ,((["(15.27)"], []), ((a ╌ b) · (c ╌ d) ≗ (a · c ⊹ b · d) ╌ (a · d ⊹ b · c)))
  ,((["(15.28)"], []), ((a ╌ b) ≗ (c ╌ d) ≡ (a ⊹ d) ≗ (b ⊹ c)))
  ,((["(15.29)"], []), ((a ╌ b) · c ≗ (a · c ╌ b · c)))
  ,((["(15.29a)"], []), ((a ⊹ b) · (a ╌ b) ≗ (a · a ╌ b · b)))

  ,((["(15.30)"], ["Axiom, Addition", "Additivity of $\\pos$"]),
      (positive @. a ∧ positive @. b ⇒ positive @. (a ⊹ b)))
  ,((["(15.31)"], ["Axiom, Multiplication", "Multiplicativity of $\\pos$"]),
      (positive @. a ∧ positive @. b ⇒ positive @. (a · b)))
  ,((["(15.32)"], ["Zero is not positive"]),
      lnot (positive @. IntLit 0))
  ,((["(15.33)"], ["Axiom, Addition", "Additivity of $\\pos$"]),
      (b ≠ IntLit 0 ⇒ positive @. b ≡ lnot (positive @. (preMinus b))))


  ,((["(15.36)"], ["Axiom, Less", "Definition of <"]), a .< b ≡ positive @. (b ╌ a))
  ,((["(15.37)"], ["Axiom, Greater", "Definition of >"]), a .> b ≡ positive @. (a ╌ b))
  ,((["(15.37.1)"], ["Greater as dual of less", "> as dual of <"]), a .> b ≡ b .< a)
  ,((["(15.38)"], ["Axiom, At most", "Definition of ≤"]), a ≤ b ≡ a .< b ∨ a ≗ b)
  ,((["(15.39)"], ["Axiom, At least", "Definition of ≥"]), a ≥ b ≡ a .> b ∨ a ≗ b)
  ,((["(15.39.1)"], ["At least as dual of at most", "≥ as dual of ≤"]), a ≥ b ≡ b ≤ a)

  ,((["(15.41)", "(15.41a)"], ["Transitivity of <", "Transitivity"]),
      (a .< b ∧ b .< c ⇒ a .< c))
  ,((["(15.41)", "(15.41b)"], ["Transitivity of ≤", "Transitivity"]),
      (a ≤ b ∧ b ≤ c ⇒ a ≤ c))
  ,((["(15.41)", "(15.41c)"], ["< followed by ≤", "Transitivity"]),
      (a .< b ∧ b ≤ c ⇒ a .< c))
  ,((["(15.41)", "(15.41d)"], ["≤ followed by <", "Transitivity"]),
      (a ≤ b ∧ b .< c ⇒ a .< c))
  ,((["(15.41)", "(15.41e)"], ["Transitivity of >", "Transitivity"]),
      (a .> b ∧ b .> c ⇒ a .> c))
  ,((["(15.41)", "(15.41f)"], ["Transitivity of ≥", "Transitivity"]),
      (a ≥ b ∧ b ≥ c ⇒ a ≥ c))
  ,((["(15.41)", "(15.41g)"], ["> followed by ≥", "Transitivity"]),
      (a .> b ∧ b ≥ c ⇒ a .> c))
  ,((["(15.41)", "(15.41h)"], ["≥ followed by >", "Transitivity"]),
      (a ≥ b ∧ b .> c ⇒ a .> c))
  ,((["(15.41.1)", "(15.41.1b)"], ["Transitivity of ≤", "Transitivity"]),
      (a ≤ b ∧ b ≤ c ≡ a ≤ b ∧ b ≤ c ∧ a ≤ c))
  ,((["(15.42)", "(15.42a)"], ["Monotonicity of + wrt. <", "Monotonicity"]),
      (a .< b ≡ a ⊹ d .< b ⊹ d))
  ,((["(15.42)", "(15.42b)"], ["Monotonicity of + wrt. ≤", "Monotonicity"]),
      (a ≤ b ≡ a ⊹ d ≤ b ⊹ d))
  ,((["(15.42)", "(15.42c)"], ["Monotonicity of + wrt. >", "Monotonicity"]),
      (a .> b ≡ a ⊹ d .> b ⊹ d))
  ,((["(15.42)", "(15.42d)"], ["Monotonicity of + wrt. ≥", "Monotonicity"]),
      (a ≥ b ≡ a ⊹ d ≥ b ⊹ d))
  ,((["(15.43)", "(15.43a)"], ["Monotonicity of · wrt. <", "Monotonicity"]),
      (IntLit 0 .< d ⇒ (a .< b ≡ a · d .< b · d)))
  ,((["(15.43)", "(15.43b)"], ["Monotonicity of · wrt. ≤", "Monotonicity"]),
      (IntLit 0 .< d ⇒ (a ≤ b ≡ a · d ≤ b · d)))
  ,((["(15.43)", "(15.43c)"], ["Monotonicity of · wrt. >", "Monotonicity"]),
      (d .> IntLit 0 ⇒ (a .> b ≡ a · d .> b · d)))
  ,((["(15.43)", "(15.43d)"], ["Monotonicity of · wrt. ≥", "Monotonicity"]),
      (d .> IntLit 0 ⇒ (a ≥ b ≡ a · d ≥ b · d)))
  ,((["(15.44)"], ["Trichotomy"]),
      (a .< b ≡ a ≗ b ≡ a .> b) ∧ lnot (a .< b ∧ a ≗ b ∧ a .> b))
  -- |,((["(15.44)", "(15.44a)"], ["Trichotomy"]), a .< b ≡ a ≗ b ≡ a .> b)|
  -- |,((["(15.44)", "(15.44b)"], ["Trichotomy"]), lnot (a .< b ∧ a ≗ b ∧ a .> b))|
  ,((["(15.44.1)"], []), a .< b ∨ a ≗ b ∨ a .> b)
  ,((["(15.44.2)"], []), lnot (a .< b ∧ a ≗ b))
  ,((["(15.44.3)"], []), a ≤ b ≡ lnot (b .< a))
  ,((["(15.44.4)"], []), a .< b ≡ lnot (b ≤ a))
  ,((["(15.45)", "(15.45a)"], ["Antisymmetry of ≤", "Antisymmetry"]),
      (a ≤ b ∧ b ≤ a ⇒ a ≗ b))
  ,((["(15.45)", "(15.45a)"], ["Antisymmetry of ≥", "Antisymmetry"]),
      (a ≥ b ∧ b ≥ a ⇒ a ≗ b))
  ,((["(15.46)", "(15.46a)"], ["Reflexivity of ≤", "Reflexivity"]),
      (a ≤ a))
  ,((["(15.46)", "(15.46a)"], ["Reflexivity of ≥", "Reflexivity"]),
      (a ≥ a))
  ,((["(15.53)", "(15.53a)"], ["Definition of ↓"]),
      (z ≤ x ↓ y ≡ z ≤ x ∧ z ≤ y))
  ,((["(15.53)", "(15.53b)"], ["Definition of ↑"]),
      (z ≥ x ↑ y ≡ z ≥ x ∧ z ≥ y))
  ,((["(15.201.2)"], ["Upto Sets"]), (x ∈ a ⋯ b ≡ a ≤ x ∧ x ≤ b))
  ,((["(15.201.4)"], []), (b ∈ a ⋯ d ∧ c ∈ b ⋯ d ≡ c ∈ a ⋯ d ∧ b ∈ a ⋯ c))
  ,((["(15.201.5)"], []), (b .< a ⇒  a ⋯ b ≗ emptyset))
  ,((["(15.201.6)"], []), (b .< a ⇒  setSize (a ⋯ b) ≗ IntLit 0))
  ]
\end{code}
%.< ≤ ≥ .>
%}}}
%}}}

%{{{ \subsection{|Show| Material}
\subsection{|Show| Material}

\begin{code}
showsSpace :: ShowS
showsSpace = (' ' :)

spaced :: ShowS -> ShowS
spaced f = showsSpace . f . showsSpace
\end{code}

\begin{code}
showsVar :: Var -> ShowS
showsVar = (++)
\end{code}

\begin{code}
showsSep, showsSep1 :: ShowS -> [ShowS] -> ShowS
showsSep1 sep = foldr1 (\ s r -> s . sep . r)
showsSep _ [] = id
showsSep sep ss = showsSep1 sep ss

showsCommaSep, showsCommaSep1 :: [ShowS] -> ShowS
showsCommaSep1 = showsSep1 (", " ++)
showsCommaSep [] = id
showsCommaSep ss = showsCommaSep1 ss
\end{code}

\begin{code}
opShows :: Op -> ShowS
opShows '𝔹' = ("Bool" ++)  -- set of Booleans
opShows 'Ⓕ' = ("false" ++)
opShows 'Ⓣ' = ("true" ++)
opShows '↙' = ("fst" ++)
opShows '↘' = ("snd" ++)
opShows '⇙' = ("Dom" ++)
opShows '⇘' = ("Ran" ++)
opShows '𝕀' = ("I" ++)
opShows '⒫' = ("pos" ++)
opShows '⁻' = ("^*" ++)  -- |relRTC|
opShows c = (c :)
\end{code}

\begin{code}
showsOp :: Op -> [ShowS] -> ShowS
showsOp '.' [f,a] = f . case a [] of
  '(' : _ -> a
  _ -> ('.' :) . a
showsOp '𝔹' [] = ("Bool" ++)  -- set of Booleans
showsOp 'Ⓕ' [] = ("false" ++)
showsOp 'Ⓣ' [] = ("true" ++)
showsOp '⋯' [] = (" .. " ++)
showsOp '↙' [] = ("fst" ++)
showsOp '↘' [] = ("snd" ++)
showsOp c@'⇙' [] = opShows c
showsOp c@'⇘' [] = opShows c
showsOp c@'𝕀' [] = opShows c
showsOp c@'˘' [s] = s . opShows c    -- |relConv|
showsOp c@'⁺' [s] = s . opShows c    -- |relTC|
showsOp c@'⁻' [s] = s . opShows c    -- |relRTC|
showsOp '¶' [s] = ("set(" ++) . s . (')' :)
showsOp 'ℙ' [s] = ("ℙ " ++) . s
showsOp '⒯' [] = ("()" ++)
-- |showsOp '⒯' [s] = error $ "showsOp: 1-tuple: '(" ++ s ")!"|
showsOp '⒯' [s] = s -- occurs as set comprehension element expression
showsOp '⒯' ss = ('(' :) . showsCommaSep1 ss . (')' :)
showsOp '|' [] = (" |" ++)
showsOp '•' [] = ("• " ++)
-- |showsOp '≔' [] = (" ≔ " ++)|
showsOp '≔' [] = (" := " ++)
-- |showsOp '≔' [x,y] = x . (" ≔ " ++) . y|
showsOp '≔' [x,y] = x . (" := " ++) . y
showsOp c@'∃' [] = opShows c
showsOp c@'∀' [] = opShows c
showsOp c@'∃' ss = case reverse ss of
  t : r : rvs@(_ : _) -> showsQuant (c :) (reverse rvs) r t
  _ -> error $ "LogicE.showsOp: incomplete existential quantification"
showsOp c@'∀' ss = case reverse ss of
  t : r : rvs@(_ : _) -> showsQuant (c :) (reverse rvs) r t
  _ -> error $ "LogicE.showsOp: incomplete universal quantification"
showsOp '⨃' ss = case reverse ss of
  t : r : rvs@(_ : _) -> showsSetCompr (reverse rvs) r (Just t)
  -- \edcomm{WK}{If we can recognise |t| as the tuple of the variables |rvs|, we can pass |Nothing| for |t|.}
  _ -> error $ "LogicE.showsOp: incomplete set comprehension"
showsOp '︷' [t] = t -- set element expression
showsOp '⇇' ss = case splitSubstArgs ss of
  Left msg -> error $ "showsOp: Internal error: " ++ msg
  -- |Right (e,(es,vs)) -> ('(' :) . e . (')' :) . showsSubst (spaced ('≔' :)) vs es|
  Right (e,(es,vs)) -> ('(' :) . e . (')' :) . showsSubst (spaced (":=" ++)) vs es
showsOp c [] = (c :)
showsOp c [x] = (c :) . x
showsOp c [x,y] = x . spaced (c :) . y
showsOp c ss = error $ "LogicE.showsOp: undefined for character " ++ charHex c ": '" ++ c : "' with " ++ shows (length ss)" arguments."
\end{code}

\begin{code}
charHex :: Char -> ShowS
charHex n = ("0x" ++) . h (Char.ord n)
  where
    h k = let (d , r) = k `divMod` 16
          in (if d == 0 then id else h d) . (hexChar r :)
    hexChar r = Char.chr (r + if r < 10 then Char.ord '0' else Char.ord 'A' - 10)
\end{code}

\begin{code}
splitSubstArgs :: [a] -> Either String (a,([a],[a]))
splitSubstArgs xs = case reverse xs of
  [] -> Left "splitSubstArgs: no term"
  e : ys -> fmap ((,) e) $ case length ys `divMod` 2 of
    (_, 1) -> Left "splitSubstArgs: substitution does not match up!"
    (k, _) -> Right $ splitAt k ys
\end{code}

\begin{code}
showsQuant :: ShowS -> [ShowS] -> ShowS -> ShowS -> ShowS
showsQuant q vs r t  = ('(' :) . q . (' ' :) . showsCommaSep1 vs . (" |" ++)
                     . r . ("• " ++) . t . (')' :)
\end{code}

\begin{code}
showsSetCompr :: [ShowS] -> ShowS -> Maybe ShowS -> ShowS
showsSetCompr vs r mt  = ('{' :) . showsCommaSep1 vs . (" |" ++)
                     . r . maybe id (("• " ++) .) mt . ('}' :)
\end{code}

\begin{code}
infixOp :: Op -> ShowS
infixOp c = spaced (c :)
\end{code}

\begin{code}
mutual :: Op -> Op -> Bool
-- |mutual '≡' '≢' = True|
-- |mutual '≢' '≡' = True|
mutual _ _ = False
\end{code}

\begin{code}
opPrec :: Op -> Int
opPrec c = case Map.lookup c opPrecs of
  Nothing -> error $ "LogicE.opPrec: undefined for undefined for character " ++ charHex c ": '" ++ c : "'"
  Just i -> i

opPrecs :: Map Op Int
opPrecs = foldr ($) Map.empty $ zipWith f [1..] (reverse opPrecTable)
  where
    f i cs m = foldr (\ c -> Map.insert c i) m cs
\end{code}

\begin{code}
opPrecTable :: [String]
opPrecTable =
  ["⇇⒯" -- substitution, tupling
  ,":" -- type annotation, not properly documented in LADM
  ,"." -- function application
  ,"˘⁺⁻" -- unary postfix
  ,"#¬∼ℙ¶︷₋₊" -- unary prefix
  ,"＾"
  ,"·/⋆"
  ,"+-∩∪×⊍∘⨾➝／＼⊕⊗⊖"
  ,"⊓⊔↑↓"
  ,"⋯"
--  |,"#"| -- \edcomm{WK}{What is this as binary operator?}
  ,"=≠<>≤≥≮≰≯≱∈∉∋∌⊂⊆⊄⊈⊃⊅⊇⊉→↔" -- conjunctional
  ,"∨∧"
   ++ "∀∃⨃" -- \edcomm{WK}{Why are these needed for |opPrec|?}
  ,"⇒⇐⇏⇍"
  ,"≡≢"
  ]
\end{code}

\begin{code}
constOps :: String
constOps = "𝔹ℕℤℚℝℂⒻⓉ∅"
\end{code}

\begin{code}
instance Show (AC.Term Var Op Const) where
  showsPrec = AC.showsPrecTerm showsVar showsConst showsOp infixOp mutual opPrec Nothing
\end{code}

\begin{code}
instance Show Expr where
  showsPrec = showsPrecExpr showsVar showsOp mutual opPrec Nothing
\end{code}

\begin{code}
showsTermHTML :: AC.Term Var Op Const -> ShowS
showsTermHTML = AC.showsPrecTerm showsVar showsConst showsOpHTML (spaced . opHTML) mutual opPrec Nothing 0
\end{code}

\begin{code}
showsExprHTML :: Expr -> ShowS
showsExprHTML = showsPrecExpr showsVar showsOpHTML mutual opPrec Nothing 0
\end{code}

%{{{ showsOpHTML :: Op -> [ShowS] -> ShowS
\begin{code}
unicodeEntity :: Char -> ShowS
unicodeEntity c = ('&' :) . ('#' :) . shows (Char.ord c) . (';' :)
\end{code}

\begin{code}
opHTML :: Op -> ShowS
opHTML '𝔹' = ("Bool" ++)  -- set of Booleans
opHTML c@'ℕ' = unicodeEntity c
opHTML c@'ℤ' = unicodeEntity c
opHTML c@'ℚ' = unicodeEntity c
opHTML c@'ℝ' = unicodeEntity c
opHTML c@'ℂ' = unicodeEntity c
opHTML 'Ⓕ' = ("<i>false</i>" ++)
opHTML 'Ⓣ' = ("<i>true</i>" ++)

opHTML '↙' = ("fst" ++)
opHTML '↘' = ("snd" ++)
opHTML '⇙' = ("Dom" ++)
opHTML '⇘' = ("Ran" ++)
opHTML '⒫' = ("pos" ++)
opHTML c@'𝕀' = unicodeEntity c
opHTML '¬' = ("&not;" ++)
opHTML c@'∼' = unicodeEntity c    -- |setCompl|
opHTML c@'˘' = unicodeEntity c    -- |relConv|
opHTML '⁺' = ("<sup>+</sup>" ++)  -- |relTC|
opHTML '⁻' = ("<sup>*</sup>" ++)  -- |relRTC|
opHTML c@'ℙ' = unicodeEntity c

opHTML c@'⨾' = unicodeEntity c
opHTML '∘' = ("&circ;" ++)
opHTML '·' = ("&sdot;" ++)
opHTML '＾' = ("**" ++)
opHTML '∩' = ("&cap;" ++)
opHTML '∪' = ("&cup;" ++)
opHTML '×' = ("&times;" ++)
opHTML '≠' = ("&ne;" ++)
opHTML '<' = ("&lt;" ++)
opHTML '>' = ("&gt;" ++)
opHTML '≤' = ("&le;" ++)
opHTML '≥' = ("&ge;" ++)
opHTML c@'≮' = unicodeEntity c
opHTML c@'≰' = unicodeEntity c
opHTML c@'≯' = unicodeEntity c
opHTML c@'≱' = unicodeEntity c
opHTML c@'➝' = unicodeEntity c
opHTML c@'／' = unicodeEntity c
opHTML c@'＼' = unicodeEntity c
opHTML '∈' = ("&isin;" ++)
opHTML '∉' = ("&notin;" ++)
opHTML '→' = ("&rarr;" ++)
opHTML '↔' = ("&lrarr;" ++)
opHTML '∨' = ("&or;" ++)
opHTML '∧' = ("&and;" ++)
opHTML '⇒' = ("&rArr;" ++)
opHTML '⇐' = ("&lArr;" ++)
opHTML c@'⇏' = unicodeEntity c
opHTML c@'⇍' = unicodeEntity c
opHTML '≡' = ("&equiv;" ++)
opHTML c@'≢' = unicodeEntity c

opHTML c@'⋆' = unicodeEntity c
opHTML '≔' = (" := " ++)

opHTML '⊓' = ("&nbsp;<b>min</b>&nbsp;" ++)
opHTML '⊔' = ("&nbsp;<b>max</b>&nbsp;" ++)
opHTML '⋯' = ("&nbsp;<b>.</b>&nbsp;<b>.</b>&nbsp;" ++)
opHTML c@'↓' = unicodeEntity c
opHTML c@'↑' = unicodeEntity c

opHTML c = if Char.ord c > 126 then unicodeEntity c else (c :)
\end{code}

\begin{code}
showsOpHTML :: Op -> [ShowS] -> ShowS
showsOpHTML '.' [f,a] = f . case a [] of
  '(' : _ -> a
  _ -> ('.' :) . a
showsOpHTML '⒯' [] = ("()" ++)
showsOpHTML '⒯' [s] = error $ "showsOpHTML: 1-tuple: '(" ++ s ")!"
showsOpHTML '⒯' ss = ('(' :) . foldr1 (\ s r -> s . (", " ++) . r) ss . (')' :)
showsOpHTML '¶' [s] = ("set(" ++) . s . (')' :)
showsOpHTML c@'˘' [s] = s . opHTML c  -- |relConv|
showsOpHTML c@'⁺' [s] = s . opHTML c  -- |relTC|
showsOpHTML c@'⁻' [s] = s . opHTML c  -- |relRTC|
showsOpHTML c@'ℙ' [s] = unicodeEntity c . (' ' :) . s
showsOpHTML '|' [] = (" |" ++)
showsOpHTML c@'•' [] = opHTML c . (' ' :)
showsOpHTML '≔' [] = (" := " ++)
showsOpHTML '≔' [x,y] = x . (" := " ++) . y
showsOpHTML c []     = opHTML c
showsOpHTML c [x]    = opHTML c . x
showsOpHTML c [x,y]  = x . spaced (opHTML c) . y
showsOpHTML c _ = error $ "LogicE.showsOpHTML: undefined for '" ++ c : "'"
\end{code}
%}}}


%{{{ showsPrecExpr
\begin{code}
showsPrecExpr ::
     (Var -> ShowS)             -- |showsVar|
  -> (Op -> [ShowS] -> ShowS)   -- |showOp|
  -> (Op -> Op -> Bool)         -- |mutual|
  -> (Op -> Int)                -- |opPrec|
  -> Maybe Op                   -- ||
  -> Int
  -> Expr
  -> ShowS

showsPrecExpr shVar shOp mut opPr = h
  where
    parens False s = s
    parens True s = ('(' :) . s . (')' :)
    needParens mOp p op p' = case compare p p' of
      GT ->  True
      EQ ->  not ((isAC op && mOp == Just op)
                  || (maybe True (mut op) mOp))
      LT ->  False
    h _    _ (Var v) = shVar v
    h _    _ (IntLit i) = shows i
    h _    _ (Const op) = shOp op []
    h mOp  p (Un op t) = let p' = opPr op in parens
      (needParens mOp p op p')
      (shOp op [h (Just op) p' t])
    h mOp  p (Bin op t t') = let p' = opPr op in parens
      (needParens mOp p op p')
      (shOp op $ map (h (Just op) p') [t, t'])
    h mOp  p (VarArgs op ts) = let p' = opPr op in parens
      (needParens mOp p op p')
      (shOp op $ map (h (Just op) p') ts)
    h mOp  p (TypeAnn e t) = h mOp  p (Bin ':' e t)
    h _mOp  _p (ConjChain _ []) = error "Internal error: showsPrecExpr: unexpected empty ConjChain"
    h mOp  p (ConjChain e ps@((op, _) : _)) = let
        p' = opPr op -- same for all chain operators
        h' = h (Just op) p'
      in parens
        (needParens mOp p op p')
        (let  k t [] = t
              k t ((o, e') : ps') = let t' =  h' e'
                    in  shOp o $ t : [k t' ps']
         in k (h' e) ps)
    h _mOp _p (Subst e vs ts) = let
        p' = opPr '⇇'
        vs' = map shVar vs
        ts' = map (h Nothing 0) ts
      in h (Just '⇇') p' e . showsSubst (shOp '≔' []) vs' ts'
    h _mOp _p (Quant op vs mr e) = parens True $ let
        p' = opPr ':'
        h' = h Nothing 0
        shRange Nothing = id
        shRange (Just r) = (' ' :) . h' r . (' ' :)
      in shOp op [] . shDecls p' vs . shOp '|' [] . shRange mr . shOp '•' [] . h' e . (' ' :)
    h _mOp _p (SetEnum ts) = ('{' :) . showsCommaSep (map (h Nothing 0) ts) . ('}' :)
    h _mOp _p (SetCompr vs mr me) = let
        p' = opPr ':'
        h' = h Nothing 0
        shRange Nothing = id
        shRange (Just r) = (' ' :) . h' r . (' ' :)
        shTerm Nothing = id
        shTerm (Just e) = shOp '•' [] . h' e
      in ('{' :) . shDecls p' vs . shOp '|' [] . shRange mr . shTerm me . ('}' :)
    shDecls p' vs = (' ' :) . showsSep1 ("; " ++) (map shDecl $ splitDecl vs)
          where
            shDecl (vs', mt) = showsCommaSep1 (map shVar vs') . case mt of
              Nothing -> id
              Just t -> (" : " ++) . h (Just ':') p' t
\end{code}

We automatically group consecutive variables with the same type:
\begin{code}
splitDecl :: [(Var, Maybe Expr)] -> [([Var],Maybe Expr)]
splitDecl = map h . groupBy (eq `on` snd)
  where
    eq Nothing   Nothing   = True
    eq (Just x)  (Just y)  = toTerm x == toTerm y
    eq _         _         = False
    h ((x, m) : ps) = (x : map fst ps, m)
    h _ = error "splitDecl: impossible!"
\end{code}

\begin{code}
showsSubst :: ShowS -> [ShowS] -> [ShowS] -> ShowS
showsSubst becomes vs es = ('[' :) . showsCommaSep1 vs . becomes . showsCommaSep1 es . (']' :)
\end{code}
%}}}

%}}}

%{{{ \subsection{Operators}
\subsection{Operators}

%{{{ isAC
|isAC| is the characteristic function on operator characters
for the predicate ``associative and commutative''.
\begin{code}
isAC :: Op -> Bool
isAC = flip elem "≡≢∧∨+·∩∪∾⊖"
\end{code}
%}}}

%{{{ isC
|isC| is the characteristic function on operator characters
for the predicate ``commutative, but not associative''.
\begin{code}
isC :: Op -> Bool
isC = flip elem "="
\end{code}
%}}}

%{{{ data CalcOpKind = COSym | COConverse Op
\begin{code}
data CalcOpKind = COSym | COConverse Op
\end{code}
%}}}

%{{{ calcOpKind :: Op -> Maybe CalcOpKind
\begin{code}
calcOpKind :: Op -> Maybe CalcOpKind
calcOpKind '=' = Just COSym
calcOpKind '≠' = Just COSym
calcOpKind '≡' = Just COSym
calcOpKind '≢' = Just COSym
calcOpKind '<' = Just (COConverse '>')
calcOpKind '>' = Just (COConverse '<')
calcOpKind '≤' = Just (COConverse '≥')
calcOpKind '≥' = Just (COConverse '≤')
calcOpKind '≮' = Just (COConverse '≯')
calcOpKind '≯' = Just (COConverse '≮')
calcOpKind '≰' = Just (COConverse '≱')
calcOpKind '≱' = Just (COConverse '≰')
calcOpKind '⇒' = Just (COConverse '⇐')
calcOpKind '⇐' = Just (COConverse '⇒')
calcOpKind '⇏' = Just (COConverse '⇍')
calcOpKind '⇍' = Just (COConverse '⇏')
calcOpKind '∈' = Just (COConverse '∋')
calcOpKind '∋' = Just (COConverse '∈')
calcOpKind '∉' = Just (COConverse '∌')
calcOpKind '∌' = Just (COConverse '∉')
calcOpKind _ = Nothing
\end{code}
%}}}

\begin{code}
isTrue :: Expr -> Bool
isTrue (Const 'Ⓣ') = True
isTrue _ = False
\end{code}

\begin{spec}
conjOp op1 op2 = Just op3  ⇒  (x `op1` y  ∧  y `op2` z ⇒ x `op3` z)
\end{spec}

\begin{code}
conjOp :: Op -> Op -> Maybe Op
conjOp '='  c    = Just c
conjOp c    '='  = Just c
conjOp '≡'  '≡'  = Just '≡'
conjOp '⇒'  '⇒'  = Just '⇒'
conjOp '⇒'  '≡'  = Just '⇒'
conjOp '≡'  '⇒'  = Just '⇒'
conjOp '⇐'  '⇐'  = Just '⇐'
conjOp '⇐'  '≡'  = Just '⇐'
conjOp '≡'  '⇐'  = Just '⇐'
conjOp '⊂' '⊂' = Just '⊂'
conjOp '⊂' '⊆' = Just '⊂'
conjOp '⊆' '⊂' = Just '⊂'
conjOp '⊆' '⊆' = Just '⊆'
conjOp '⊃' '⊃' = Just '⊃'
conjOp '⊇' '⊃' = Just '⊃'
conjOp '⊃' '⊇' = Just '⊃'
conjOp '⊇' '⊇' = Just '⊇'
conjOp _    _     = Nothing
\end{code}

\begin{code}
conjOps :: [Op] -> Maybe Op
conjOps [] = Nothing
conjOps (op : ops) = foldl' (\ m o -> do o' <- m; conjOp o' o) (Just op) ops
\end{code}

%{{{ Haskell precedences
The Haskell precedences follow \cite{Gries-Schneider-1993} ---
we haven't implemented all operators yet,
and Haskell has only ten precedence levels.

\begin{code}
infixr 9 @.
infixr 8 ＾
infixl 7 ·, ÷
infixl 6 ⊹, ╌, ∩, ∪, ×
infixr 5 →, ↔,⋯
infixl 4 ∶
infixl 3 ≗, ≠, .>, .<, ≮, ≯, ≤, ≥, ≰, ≱, ∈, ∉, ⊂, ⊆, ⊄, ⊈, ⊃, ⊅, ⊇, ⊉
infixl 2 ∧, ∨
infixl 1 ⇐, ⇍
infixr 1 ⇒, ⇏
infixr 0 ≡, ≢
\end{code}
%}}}

\begin{code}
(≡), (≢), (⇒), (⇐), (⇏), (⇍), (∧), (∨) :: Expr -> Expr -> Expr
(≡) = Bin '≡'
(≢) = Bin '≢'
(⇒) = Bin '⇒'
(⇐) = Bin '⇐'
(⇏) = Bin '⇏'
(⇍) = Bin '⇍'
(∧) = Bin '∧'
(∨) = Bin '∨'
\end{code}

\begin{code}
(⊂), (⊆), (⊄), (⊈), (⊃), (⊅), (⊇), (⊉) :: Expr -> Expr -> Expr
(⊂) = Bin '⊂'
(⊆) = Bin '⊆'
(⊄) = Bin '⊄'
-- x ⊄ y = lnot (x ⊂ y)
(⊈) = Bin '⊈'
-- x ⊈ y = lnot (x ⊆ y)
(⊃) = Bin '⊃'
(⊅) = Bin '⊅'
(⊇) = Bin '⊇'
(⊉) = Bin '⊉'
\end{code}

\begin{code}
tuple :: [Expr] -> Expr
tuple = VarArgs '⒯'

pair :: Expr -> Expr -> Expr
pair x y = tuple [x,y]
\end{code}

\begin{code}
(@.), (⨾), (∘), (≗), (≠), (.>), (.<), (≮), (≯), (≤), (≥), (≰), (≱)
  , (∈), (∉), (∋), (∌), (➝), (／), (＼) :: Expr -> Expr -> Expr
(@.) = Bin '.'
(⨾) = Bin '⨾'
(∘) = Bin '∘'
(≗) = Bin '='
(≠) = Bin '≠'
-- x ≠ y = lnot (x ≗ y)
(.<) = Bin '<'
(.>) = Bin '>'
(≮) = Bin '≮'
(≯) = Bin '≯'
(≤) = Bin '≤'
(≥) = Bin '≥'
(≰) = Bin '≰'
(≱) = Bin '≱'
(∈) = Bin '∈'
(∉) = Bin '∉'
-- x ∉ s = lnot (x ∈ s)
(∋) = Bin '∋'
(∌) = Bin '∌'
(➝) = Bin '➝'  -- relative pseudo-complement
(／) = Bin '／' -- left residual
(＼) = Bin '＼' -- right residual
\end{code}

\begin{code}
(·), (＾), (÷), (⊹), (╌), (∩), (∪), (×), (→), (↔),(∶), (⋆) :: Expr -> Expr -> Expr
(·) = Bin '·'
(＾) = Bin '＾'
(÷) = Bin '/'
(⊹) = Bin '+'
(╌) = Bin '-'
(∩) = Bin '∩'
(∪) = Bin '∪'
(×) = Bin '×'
(→) = Bin '→'
(↔) = Bin '↔'
(∶) = Bin ':'
(⋆) = Bin '⋆'
\end{code}

\begin{code}
(↓), (↑), (⊓), (⊔), (⊗), (⊕), (⊖), (⋯) :: Expr -> Expr -> Expr
(↓) = Bin '↓'
(↑) = Bin '↑'
(⊓) = Bin '⊓'
(⊔) = Bin '⊔'
(⊗) = Bin '⊗'
(⊕) = Bin '⊕'
(⊖) = Bin '⊖'
(⋯) = Bin '⋯'
\end{code}

\begin{code}
lnot, setCompl :: Expr -> Expr
lnot = Un '¬'
setCompl = Un '∼'

universe :: Expr
universe = Const '𝐔'
\end{code}

\begin{code}
setSize :: Expr -> Expr
setSize = Un '#'
\end{code}

\begin{code}
power, setType :: Expr -> Expr
power = Un 'ℙ'
setType = Un '¶'
\end{code}

\begin{code}
relConv, relTC, relRTC :: Expr -> Expr
relConv = Un '˘'
relTC = Un '⁺'
relRTC = Un '⁻'
\end{code}

\begin{code}
relDom, relRan, relId :: Expr
relDom = Const '⇙'
relRan = Const '⇘'
relId = Const '𝕀'
\end{code}

\begin{code}
positive :: Expr
positive = Const '⒫'
\end{code}

\begin{code}
prePlus, preMinus :: Expr -> Expr
prePlus = Un '₊'
preMinus = Un '₋'
\end{code}

\begin{code}
false, true :: Expr
false  = Const 'Ⓕ'
true   = Const 'Ⓣ'
\end{code}

\begin{code}
bool, nat, integer, rational, real, complex, emptyset :: Expr
bool      = Const '𝔹'   -- could not find $\mathbb{B}$!
nat       = Const 'ℕ'
integer   = Const 'ℤ'
rational  = Const 'ℚ'
real      = Const 'ℝ'
complex   = Const 'ℂ'
emptyset  = Const '∅'

numTypes, fractionalNumTypes, realNumTypes, rationalNumTypes, integralNumTypes :: [Expr]
numTypes = [nat, integer, rational, real, complex]
fractionalNumTypes = [rational, real, complex]
realNumTypes = [nat, integer, rational, real]
rationalNumTypes = [nat, integer, rational]
integralNumTypes = [nat, integer]
\end{code}

Pair projections:
\begin{code}
proj1, proj2 :: Expr
proj1 = Const '↙'
proj2 = Const '↘'
\end{code}

\begin{code}
quantOps :: [(Op, Op)]
quantOps =
  [ ('≡','≣')
  , ('·','∏')
  , ('+','∑')
  , ('∩','⋂')
  , ('∪','⋃')
  , ('↓','⇊')
  , ('↑','⇈')
  , ('∧','∀')
  , ('∨','∃')
  , ('⋆','★') -- metavariable for quantification operators in \cite{Gries-Schneider-1993}
  , ('⊍','⨃') -- set comprehension
  ]
\end{code}

\begin{code}
binToQuant, quantToBin :: Op -> Maybe Op
binToQuant =  let mb = Map.fromList quantOps
              in flip Map.lookup mb
quantToBin =  let mq = foldr (\ (b,q) -> Map.insert q b) Map.empty quantOps
              in flip Map.lookup mq
\end{code}

%}}}

%{{{ \subsection{Evaluation}
\subsection{Evaluation}

\begin{code}
checkLaw :: Theorem -> IO ()
checkLaw (([], ds),e) = putStrLn $ "Law without number?" ++ unwords (ds ++ [show e])
checkLaw th@(_, e) = case isValid e of
    Right True -> -- return ()
                  putStrLn $ "OK: " ++ name
    Right False ->  putStrLn $ "FAILED: " ++ name
    Left err -> putStrLn $ "While checking " ++ name ++ ": " ++ err
  where name = showTheorem th

checkLaws :: IO ()
checkLaws = mapM_ checkLaw laws
\end{code}

\begin{code}
isValid :: (Monad m) => Expr -> m Bool
isValid e = let
    vs = Set.toList $ freeVars e
    envs = h vs
      where
        h [] = [Map.empty]
        h (x : xs) = let ms = h xs
           in map (Map.insert x True) ms ++ map (Map.insert x False) ms
  in liftM and $ mapM (flip boolEval e) envs
\end{code}

\begin{code}
boolEval :: (Monad m) => Map Var Bool -> Expr -> m Bool
boolEval env (Bin op t1 t2) = case boolBinOp op of
  Just f -> liftM2 f (boolEval env t1) (boolEval env t2)
  Nothing -> fail $ "boolBinOp: `" ++ op : "' not found."
boolEval env (Un op t) = case boolUnOp op of
  Just f -> liftM f $ boolEval env t
  Nothing -> fail $ "boolUnOp: `" ++ op : "' not found."
boolEval _env (IntLit _) = fail $ "boolEval: unexpected IntLit"
boolEval _env (Const op) = case boolConst op of
  Just f -> return f
  Nothing -> fail $ "boolConst: `" ++ op : "' not found."
boolEval env (Var v) = case Map.lookup v env of
  Just x -> return x
  Nothing  -> fail $ "boolEval: `" ++ v ++ "' not found."
boolEval _ (VarArgs _ _) = fail $ "boolEval: unexpected VarArgs"
boolEval env (TypeAnn e _t) = boolEval env e
boolEval env (Subst e vs es) = do
  bs <- mapM (boolEval env) es
  let env' = foldr ($) env $ zipWith Map.insert vs bs
  boolEval env' e
boolEval env (ConjChain e ps) = liftM and
   $ mapM (\ (o,(t1,t2)) -> boolEval env (Bin o t1 t2)) $ unChain e ps
boolEval _ (Quant _ _ _ _) = fail $ "boolEval: unexpected quantification"
boolEval _ (SetEnum _) = fail $ "boolEval: unexpected set enumeration"
boolEval _ (SetCompr _ _ _) = fail $ "boolEval: unexpected set comprehension"
\end{code}

\begin{code}
boolConst :: Op -> Maybe Bool
boolConst 'Ⓕ' = Just False
boolConst 'Ⓣ' = Just True
boolConst _    = Nothing
\end{code}
  
\begin{code}
boolUnOp :: Op -> Maybe (Bool -> Bool)
boolUnOp '¬' = Just not
boolUnOp _ = Nothing
\end{code}
  
\begin{code}
boolBinOp :: Op -> Maybe (Bool -> Bool -> Bool)
boolBinOp c = Map.lookup c boolBinOps

boolBinOps :: Map Op (Bool -> Bool -> Bool)
boolBinOps = Map.fromList
  [('≡', (==))
  ,('≢', (/=))
  ,('⇒', (<=))
  ,('⇐', (>=))
  ,('⇏', \ p q -> not (p <= q))
  ,('⇍', \ p q -> not (p >= q))
  ,('∧', (&&))
  ,('∨', (||))
  ,('=', (==))
  ,('≠', (/=))
  ]
\end{code}
%}}}

%{{{ \subsection{Types}
\subsection{Types}

\begin{code}
α, β, γ, δ, τ :: Expr
α = Var "α"
β = Var "β"
γ = Var "γ"
δ = Var "δ"
τ = Var "τ"
\end{code}

\begin{code}
binOpType :: Expr -> Expr
binOpType a = a × a → a
\end{code}

%{{{ opTypes :: Op -> [Expr]
\begin{code}
opTypes :: Op -> [Expr]
-- |opTypes '⇇' = | -- substitution
-- |opTypes '⒯' = | -- |VarArgs| tuple
opTypes '.' = [(α → β) × α → β]
opTypes '#' = [setType α → nat]
opTypes '¬' = [bool → bool]
opTypes 'ℙ' = [setType α → setType (setType α)]
opTypes '¶' = [setType α → setType (setType α)]
opTypes '·' = map binOpType numTypes
opTypes '＾' = map binOpType numTypes
opTypes '/' = map binOpType fractionalNumTypes
opTypes '⋆' = [binOpType τ]
opTypes '+' = map binOpType numTypes
opTypes '-' = binOpType (setType α) : map binOpType numTypes
opTypes '∩' = [binOpType (setType α)]
opTypes '∪' = [binOpType (setType α)]
opTypes '↓' = map binOpType realNumTypes
opTypes '↑' = map binOpType realNumTypes
opTypes '⊓' = map binOpType realNumTypes
opTypes '⊔' = map binOpType realNumTypes
opTypes '×' = [setType α × setType β → setType (α × β)]
opTypes '⊍' = [binOpType (setType α)]
opTypes '=' = [α × α → bool]
opTypes '≠' = [α × α → bool]
opTypes '<' = map binOpType realNumTypes
opTypes '>' = map binOpType realNumTypes
opTypes '≤' = map binOpType realNumTypes
opTypes '≥' = map binOpType realNumTypes
opTypes '≮' = map binOpType realNumTypes
opTypes '≰' = map binOpType realNumTypes
opTypes '≯' = map binOpType realNumTypes
opTypes '≱' = map binOpType realNumTypes
opTypes '∈' = [α × setType α → bool]
opTypes '∉' = [α × setType α → bool]
opTypes '∋' = [setType α × α → bool]
opTypes '∌' = [setType α × α → bool]
opTypes '⊂' = [setType α × setType α → bool]
opTypes '⊆' = [setType α × setType α → bool]
opTypes '⊄' = [setType α × setType α → bool]
opTypes '⊈' = [setType α × setType α → bool]
opTypes '⊃' = [setType α × setType α → bool]
opTypes '⊅' = [setType α × setType α → bool]
opTypes '⊇' = [setType α × setType α → bool]
opTypes '⊉' = [setType α × setType α → bool]
opTypes '→' = [setType α × setType β → setType (α → β)]
opTypes '∨' = [binOpType bool]
opTypes '∧' = [binOpType bool]
opTypes ':' = [α × setType α → α]
opTypes '⇒' = [binOpType bool]
opTypes '⇐' = [binOpType bool]
opTypes '⇏' = [binOpType bool]
opTypes '⇍' = [binOpType bool]
opTypes '≡' = [binOpType bool]
opTypes '≢' = [binOpType bool]
opTypes '∅' = [setType α]
opTypes '𝔹' = [setType bool]
opTypes 'ℕ' = map setType numTypes
opTypes 'ℤ' = map setType [integer, rational, real, complex]
opTypes 'ℚ' = map setType [rational, real, complex]
opTypes 'ℝ' = map setType [real, complex]
opTypes 'ℂ' = [setType complex]
opTypes 'Ⓕ' = [bool]
opTypes 'Ⓣ' = [bool]
opTypes c = error $ "opTypes '" ++ c : "': undefined!"
\end{code}
%}}}
%}}}

%{{{ EMACS lv
% Local Variables:
% folded-file: t
% fold-internal-margins: 0
% eval: (fold-set-marks "%{{{ " "%}}}")
% eval: (fold-whole-buffer)
% end:
%}}}
