% TeXParse.lhs
%
% This file is part of CalcCheck
%
% Copyright (c) 2011  Wolfram Kahl
% All rights reserved
%
\section{|TeXParse|: Parsing Basic {TeX{}}}

\begin{code}
module TeXParse where

import TeX

import Text.ParserCombinators.Parsec hiding (Parser)

import Data.Char (isLetter)
\end{code}

\begin{code}
type Parser st = GenParser Char st
\end{code}

\begin{code}
char0 :: Char -> Parser st String
char0 c = char c >> return [c]
\end{code}

\begin{code}
(<$>) :: (a -> b) -> Parser st a -> Parser st b
f <$> p = fmap f p
\end{code}

\begin{code}
(>*>) :: Parser st [a] -> Parser st [a] -> Parser st [a]
p >*> q = do
  xs <- p
  ys <- q
  return $ xs ++ ys
\end{code}

\begin{code}
(<*) :: Parser st a -> Parser st b -> Parser st a
p <* q = do
  x <- p
  _ <- q
  return x
\end{code}


\begin{code}
pTeX :: Parser st TeX
pTeX =  (TChar <$> noneOf "%\\{}")
   <|>  (char '%' >> line >> pTeX)
-- |   <>  ((uncurry $ uncurry TMath) <$> pTeXMathWrap (many pTeX))|
   <|>  (TMacro <$> pMacro)
   <|>  (TGroup <$> brace (many pTeX))
\end{code}

\begin{code}
pTeXMathWrap :: Parser st a -> Parser st ((String, String), a)
pTeXMathWrap p =
    between (char '$' >> mspace) (mspace >> char '$') (fmap ((,) ("$","$")) p)
    <|>
    between (macro' "(" >> mspace) (mspace >> macro ")") (fmap ((,) ("\\(","\\)")) p)

showsMathWrap :: (a -> ShowS) -> ((String, String), a) -> ShowS
showsMathWrap showsA ((open,close), a) = (open ++) . showsA a . (close ++)
\end{code}

\begin{code}
mbox, fbox :: Parser st [TeX]
mbox = macro1 "mbox" (many pTeX) <?> "\\mbox{...}"
fbox = macro1 "fbox" (many pTeX) <?> "\\fbox{...}"
\end{code}

\begin{code}
rawLine :: Parser st String
rawLine = many (noneOf "\r\n")

crlf0 :: Parser st String
crlf0 = crlf >> return "\n"

crlf :: Parser st ()
crlf = (cr >> optional lf) <|> (lf >> return ())
  where  cr = char '\r'
         lf = char '\n'

line0 :: Parser st String
line0 = do
    s <- rawLine
    lf <- if null s
      then crlf0
      else option "" crlf0
    return $ s ++ lf
  <?> "<ignored line>"

line :: Parser st ()
line = line0 >> return ()
\end{code}

\begin{code}
pMacro, pMacro0 :: Parser st String
pMacro = try (char '\\' >> (((: []) <$> satisfy (not . isLetter)) <|> many1 letter))
pMacro0 = fmap ('\\' :) pMacro
\end{code}

\begin{code}
macro0, macro0' :: String -> Parser st String
macro0 s = macro s >> return ('\\' : s)
macro0' s = macro' s >> return ('\\' : s)
\end{code}

\begin{code}
macro, macro' :: String -> Parser st ()
macro s = try (string s' >> case s of
                   [c] | not (isLetter c) -> return ()
                   _ -> notFollowedBy letter)
    <?> s'
  where s' = '\\' : s
macro' s = macro s >> mspace
\end{code}

|charMacro| must be used only with non-letter characters.
\begin{code}
charMacro0 :: Char -> Parser st String
charMacro0 c = try (string s) <?> s
  where s = ['\\' , c]

charMacro :: Char -> Parser st ()
charMacro c = charMacro0 c >> return ()
\end{code}

|linebreak| simplistically parses hard line breaks
\verb|\\| including spacing variants like \verb|\\[3ex]|.

\begin{code}
linebreak0 :: Parser st String
linebreak0 = try (string crcr) >> option crcr dim
  where
    crcr = "\\\\"
    dim = do
      len <- bracket texLength
      return $ crcr ++ '[' : len ++ "]"

linebreak :: Parser st ()
linebreak = linebreak0 >> return ()
\end{code}

|texspace| is for general spacing.
\begin{code}
rawspace, texspace0, texspace :: Parser st String
rawspace  =    (choice (map char0 " ~\t\n\r\f")) <|> linebreak0
          <|>  (char0 '%' >*> line0)
          <|>  charMacro0 ' '
          <|>  kern
          <|>  choice (map macro0 ["strut", "hfill"])
          <|>  try (beamerMacro0 "onslide")
          <|>  try (macro1 "phantom" (many1 (satisfy (`notElem` "{}"))) >> return "\\phantom{}")

kern :: Parser st String
kern = try (do
  k <- string "\\kern"
  len <- texLength
  return $ k ++ len
 ) <?> "\\kern"

texLength :: Parser st String
texLength = try (do
  m <- option "" $ fmap (: []) $ char '-'
  f <- float
  u <- texUnit
  return $ m ++ f ++ u
 ) <?> "length"

texUnit :: Parser st String
texUnit = do
  u <- choice (map (try . string) ["pt","em","ex","mm","cm","\\textwidth","\\textheight"])
  notFollowedBy letter
  return u

float :: Parser st String
float = do
    d <- digit
    ds <- many digit
    f <- option "" frac
    return $ d : ds ++ f
  <|>
    frac
  where
    frac = do
      dot <- char '.'
      fs <- many digit
      return $ dot : fs

texspace0 = fmap concat $ many (rawspace <?> "<spacing>")
texspace = fmap concat $ many1 (rawspace <?> "<spacing>")
\end{code}

|mspace| is for whitespace in parsed math environments
(\verb|calc| and \verb|defs|)
that is irrelevant for the mathematical content:

\begin{code}
mspace :: Parser st ()
mspace = many  ((    (rawspace >> return ())
                <|>  choice  (map charMacro ",;!" ++ color : try (wrap mspace) :
                              map macro ["BREAK", "quad", "qquad"]
                             )) <?> "<spacing>")  >> return ()
\end{code}

To avoid problems with my slides,
we consider \verb+\color{+\emph{color name}\verb+}+ declarations
as white-space, too.
\edcomm{WK}{Only if \verb+--beamer+ was given!}

\begin{code}
betweenCC :: Parser st Char -> Parser st Char -> Parser st String -> Parser st String
betweenCC l r p = do
  c <- l
  s <- p
  d <- r
  return $ c : s ++ [d]

(>>++) :: Parser st String -> Parser st String -> Parser st String
p >>++ q = do
  xs <- p
  ys <- q
  return $ xs ++ ys

beamerRange0 :: Parser st String
beamerRange0 = betweenCC (char '<') (char '>') $ do
  l <- fmap (: []) (char '+') <|> many digit
  m <- char '-'
  r <- many digit
  return $ l ++ m : r

beamerRange :: Parser st ()
beamerRange = between (char '<') (char '>') $ do
  _ <- fmap (: []) (char '+') <|> many digit
  _ <- char '-'
  _ <- many digit
  return ()

beamerMacro0 :: String -> Parser st String
beamerMacro0 s = macro0 s >>++ option "" beamerRange0

beamerMacro :: String -> Parser st ()
beamerMacro s = macro s >> option () beamerRange

beamerMacro1 :: String -> Parser st a -> Parser st a
beamerMacro1 s p = between (beamerMacro s) texspace0 (brace p)
\end{code}

\begin{code}
color :: Parser st ()
color = try (beamerMacro1 "color" (many1 alphaNum >> return ()))

only :: Parser st ()
only = beamerMacro "only"
\end{code}

To enable coloured parentheses,
we allow braces around delimiters:

\begin{code}
onlyWrap :: Parser st a -> Parser st a
onlyWrap w = option () (only <|> macro "mathop") >> brace' w

wrap :: Parser st a -> Parser st a
wrap w = onlyWrap (between mspace mspace w)
wrapped :: Parser st a -> Parser st a
wrapped p = try w where w = p <|> wrap w
\end{code}

\begin{code}
brace, brace', bracket, paren, paren', mparen :: Parser st a -> Parser st a
brace    = between (char '{') (char '}')
brace'    = between (char '{' >> mspace) (char '}' >> mspace)
bracket  = between (char '[') (char ']' >> mspace)
paren    = between (wrapped (char '(')) (wrapped (char ')'))
paren'    = between (wrapped (char '(') >> mspace) (wrapped (char ')') >> mspace)
mparen p = let
   modParen l r = between  (try (wrapped (macro l >> char '(')) >> mspace)
                           ((let s = '\\' : r ++ ")" in wrapped (string s) <?> s) >> mspace)
  in      modParen "left" "right" p
     <|>  modParen "big" "big" p
     <|>  paren' p
  <?> "("
\end{code}

|commasep1 p| parses a non-empty comma-separated list of |p|s.
\begin{code}
commasep, commasep1, semisep1 :: Parser st a -> Parser st [a]
commasep1 p = p `sepBy1` (char ',' >> mspace)
commasep = option [] . commasep1
semisep1 p = p `sepBy1` (char ';' >> mspace)
\end{code}

|macro1 s p| parses macro |s| with argument (currently only in braces) |p|.
\begin{code}
macro1 :: String -> Parser st a -> Parser st a
macro1 s p = between (macro s) texspace0 (brace p)
\end{code}

\begin{code}
smacro1 :: String -> Parser st String -> Parser st String
smacro1 s p = fmap (('\\' :) . (s ++) . ('{' :) . (++ "}")) $ between (macro s) mspace (brace p)
\end{code}

|macro1O1 s p p'| parses macro |s| with one normal argument (currently only in braces) |p| and one optional argument in brackets |p'|.
\begin{code}
macro1O1 :: String -> Parser st a -> Parser st b -> Parser st (a, Maybe b)
macro1O1 s p p' = do
  r <- try $ macro s >> brace p
  opt <- (Just <$> bracket p') <|> (mspace >> return Nothing)
  return (r, opt)
\end{code}

|env s p| parses environment |s| with contents |p|.
\begin{code}
env :: String -> Parser st a -> Parser st a
env s = let ps = string s in between  ((try $ macro1 "begin" ps) <?> "\\begin{" ++ s ++ "}")
                                      (macro1 "end" ps <?> "\\end{" ++ s ++ "}")
\end{code}

|envO1 s opt p | parses environment |s| with optional argument |opt| and contents |p|.
\begin{code}
envO1 :: String -> Parser st opt -> Parser st a -> Parser st (Maybe opt, a)
envO1 s opt p = let ps = string s in do
  (_, mopt) <- macro1O1 "begin" ps opt <?> "\\begin{" ++ s ++ "}"
  r <- p
  _ <- macro1 "end" ps <?> "\\end{" ++ s ++ "}"
  return (mopt, r)
\end{code}

\begin{code}
rawThNum, parenThNum :: Parser st String
rawThNum = many1 (alphaNum <|> char '.')
parenThNum = paren rawThNum <?> "(<theorem number>)"
\end{code}

%{{{ EMACS lv
% Local Variables:
% folded-file: t
% fold-internal-margins: 0
% eval: (fold-set-marks "%{{{ " "%}}}")
% eval: (fold-whole-buffer)
% end:
%}}}
