\section{|Yesod|: Framework for a Web-Based Implementation of CalcCheck}
\subsection{Reading About Yesod}
A good first reading on Yesod is the O'Reilly Haskell and Yesod Book, available for free online: http://www.yesodweb.com/book.

Major updates have been made since publication, making the official Yesod Blog is a good first place to look if what you're reading seems like it may no longer be applicable: http://www.yesodweb.com/blog.

\subsection{Reason for Choosing Yesod}
There are a number of reasons Yesod was chosen for the web-based implementation of CalcCheck over other Haskell web frameworks like Snap or Happstack:
\begin{itemize}
\item It uses the Warp backend, which is currently the fastest Haskell web server;
\item It has a major focus on type safety;
\item It has a highly modularized design;
\item It is well-documented in comparison to other newer frameworks; and
\item It appears to be the most actively updated framework.
\end{itemize}

Nonetheless, other frameworks could have been used equally effectively.

\subsection{Scaffolding}
This product of this stand-alone file is an exception to the norm of Yesod sites. It doesn't use much more than the basic necessities to create a very minimal page. For working web sites, it is highly recommended by the developers of Yesod that you work from a scaffolded site.\footnote{http://www.yesodweb.com/book/scaffolding-and-the-site-template}

A scaffolded site is created using the "yesod init" command. It produces a large file structure from a fully cabalized Haskell package. GET and POST functions go in Handler files. Routes go in config/routes. Modules that are added need to be added to the .cabal file in the exported-modules list. It is also worth noting that Prelude must be explicitly imported into each module.

The "yesod dev" command launches the development site, and runs until it is manually terminated so that it can automatically rebuild and reload when code is changed. It will also print any messages related to errors that are encountered throughout testing.

\subsection{Code}
\begin{code}
{-# LANGUAGE TypeFamilies #-}
{-# LANGUAGE QuasiQuotes #-}
{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE TemplateHaskell #-}
{-# LANGUAGE OverloadedStrings #-}

module Main (main) where

import Yesod hiding (defaultLayout)
import Control.Applicative ((<$>), (<*>))

-- Used for string conversion
import Data.Text as T (unpack)
import Data.ByteString.Lazy as B (unpack)
import Codec.Binary.UTF8.String (decode)

-- Converts String (containing raw HTML) to Markup
import Text.Blaze.Internal (preEscapedString)

-- CalcCheck
import Parser (pFile)
import LogicE (showsItemsHTML)
import CheckCalcs (checkItems)
import Text.Parsec.Prim (runP)
\end{code}


mkYesod is a Template Haskell function.
RootR listens for requests to "/" and answers GET requests, and CalcR listens for requests to "/CalcCheck" and answers POST requests.\footnote{http://www.yesodweb.com/book/routing-and-handlers}
\begin{code}
data CalcCheck = CalcCheck

mkYesod "CalcCheck" [parseRoutes|
/ RootR GET
/CalcCheck CalcR POST
|]

instance Yesod CalcCheck

instance RenderMessage CalcCheck FormMessage where
    renderMessage _ _ = defaultFormMessage
\end{code}

defaultLayout contains a site layout.

In most cases the parameter "contents" will be of the form "[whamlet|<p>Some HTML</p>|]". The whamlet quasi-quote converts the contents to a Widget and has special syntax for linking, function usage/variable interpolation, etc.\footnote{http://yesodweb.com/book/shakespearean-templates}

PageContent accepts a "title" parameter, but it is not used. Instead, "CalcCheck" is the title for all pages using defaultLayout. 

The Cassius quasi-quote uses syntax very similar to the CSS template language, but with whitespace rules instead of braces and semicolons to delimit blocks.\footnotemark[\value{footnote}]
\begin{code}
defaultLayout contents = do
    PageContent title headTags bodyTags <- widgetToPageContent $ do
        toWidget [cassius|
            p
                color: #666666
            h1
                color: #660000
            |]
        addWidget contents
    hamletToRepHtml [hamlet|
        $doctype 5
        <html>
             <head>
                  <title>CalcCheck
                  ^{headTags}
             <body>
                  <h1>CalcCheck
                  <div id="wrapper">
                  ^{bodyTags}
    |]
\end{code}

The wrapper data type, File, to be received from the form is unnecessary with its current usage, but if other fields are added to the form in the future, it will be useful. E.g. An additional field might export "name :: String".

fileForm declares the form.
\begin{code}
data File = File
    { info :: FileInfo }
    deriving Show

fileForm :: Html -> MForm CalcCheck CalcCheck (FormResult File, Widget)
fileForm = renderDivs $ File
    <$> fileAFormReq ""
\end{code}
getRootR generates the form to display when it receives a request.
\begin{code}
getRootR :: Handler RepHtml
getRootR = do
    ((_, widget), enctype) <- runFormPost fileForm
    defaultLayout $ do
        [whamlet|
        <form method=post action=@{CalcR} enctype=#{enctype}>
            ^{widget}
            <input type=submit value="Check">
        |]
\end{code}

postCalcR processes the form after the Submit button is clicked.

One line of importance is:
\begin{code}
return $ runP pFile () (extractName file) (extractContents file)
\end{code}
where "runP" is a Parsec function that runs the string parser "pFile" (from the Parser module) on the file contents.

If the parser succeeds, then "checkItems" (from the CheckCalcs module) is run on the result, producing a list of items. The list is converted to a String, which preEscapedString turns into Hamlet.

Note also that the whamlet quasi-quote can accept either standard HTML syntax (which it does in our case), or a special Hamlet\footnote{http://yesodweb.com/book/shakespearean-templates} syntax.

If the form result is bad (result is of type "FormMissing" or "FormFailure [String]"), or the parsing fails, an error message is displayed and a new form is displayed for the user to try again.
\begin{code}
postCalcR :: Handler RepHtml
postCalcR = do
    ((result, widget), enctype) <- runFormPost fileForm
    case result of
      
        -- For good files submitted to form
        FormSuccess file -> do
            ecss <- return $ runP pFile () (extractName file) (extractContents file)
            case ecss of
              
                -- If parsing fails, print error and new form to try again 
                Left e ->  defaultLayout $ do
                           [whamlet| #{show e} |]
                           [whamlet|
                           <form method=post action=@{CalcR} enctype=#{enctype}>
                               ^{widget}
                               <input type=submit value="Check">
                           |]

                -- If parsing succeeds, show checked items as HTML
                Right css -> defaultLayout $ do
                             [whamlet| <p><a href=@{RootR}>Go Back</a> |]
                             [whamlet| #{preEscapedString $ showsItemsHTML (checkItems css) ""} |]
                             
        -- For bad files submitted to form
        _ -> defaultLayout [whamlet|
            <p>Invalid input. Try again...
            <form method=post enctype=#{enctype}>
                ^{widget}
                <input type=submit value="Check">
            |]
\end{code}

extractContents gets the String file contents from the form return type, and extractName gets the String file name.
\begin{code}
extractContents :: File -> String
extractContents = decode . B.unpack . fileContent . info

extractName :: File -> String
extractName = T.unpack . fileName . info
\end{code}

The main function launches the site on localhost:3000.
\begin{code}
main :: IO ()
main = warpDebug 3000 CalcCheck
\end{code}
